@@ -6,7 +6,7 @@
 #ifndef _IP_VS_H
 #define _IP_VS_H
 
-#define IP_VS_VERSION_CODE            0x000001
+#define IP_VS_VERSION_CODE            0x000003
 #define NVERSION(version)                       \
         (version >> 16) & 0xFF,                 \
         (version >> 8) & 0xFF,                  \
@@ -66,7 +66,7 @@ struct ip_vs_rule_user {
  */
 #define MASQUERADE_EXPIRE_ICMP      125*HZ
 
-#define IP_VS_BASE_CTL  	80	    /* base for IPVS socket options */
+#define IP_VS_BASE_CTL  	128	    /* base for IPVS socket options */
 
 #define IP_VS_SO_SET_NONE	IP_VS_BASE_CTL	         /* just peek */
 #define IP_VS_SO_SET_INSERT	(IP_VS_BASE_CTL+1)
@@ -84,24 +84,18 @@ struct ip_vs_rule_user {
 #define IP_VS_SO_GET_MAX	IP_VS_SO_GET_VERSION
 
 
+#define IP_VS_CONN_F_FWD_MASK         0x0003    /* mask for the fwd methods */
+#define IP_VS_CONN_F_MASQ	      0x0000    /* masquerading */
+#define IP_VS_CONN_F_LOCALNODE	      0x0001    /* local node */
+#define IP_VS_CONN_F_TUNNEL	      0x0002    /* tunneling */
+#define IP_VS_CONN_F_DROUTE           0x0003    /* direct routing */
+#define IP_VS_CONN_F_HASHED	      0x0040 	/* hashed entry */
+#define IP_VS_CONN_F_NOOUTPUT         0x0080    /* no output packets */
+#define IP_VS_CONN_F_INACTIVE         0x0100    /* not established */
+#define IP_VS_CONN_F_OUT_SEQ          0x0200    /* must do output seq adjust */
+#define IP_VS_CONN_F_IN_SEQ           0x0400    /* must do input seq adjust */
+#define IP_VS_CONN_F_NO_CPORT         0x0800    /* no cport set yet */
 
-#define IP_VS_CONN_F_NO_DADDR	      0x0001 	/* no daddr yet */
-#define IP_VS_CONN_F_NO_DPORT     	      0x0002	/* no dport set yet */
-#define IP_VS_CONN_F_NO_SADDR	      0x0004	/* no sport set yet */
-#define IP_VS_CONN_F_NO_SPORT	      0x0008	/* no sport set yet */
-
-#define IP_VS_CONN_F_DLOOSE	      0x0010	/* loose dest binding */
-#define IP_VS_CONN_F_NO_REPLY	      0x0080	/* no reply yet from outside */
-
-#define IP_VS_CONN_F_HASHED	      0x0100 	/* hashed entry */
-#define IP_VS_CONN_F_VS		  0x00010000	/* virtual server releated */
-#define IP_VS_CONN_F_NOOUTPUT     0x00020000	/* output packets */
-#define IP_VS_CONN_F_INACTIVE     0x00040000    /* not established */
-#define IP_VS_CONN_F_FWD_MASK	  0x00700000    /* mask for the fdw method */
-#define IP_VS_CONN_F_LOCALNODE	  0x00100000    /* local node destination */
-#define IP_VS_CONN_F_TUNNEL	  0x00200000    /* packets will be tunneled */
-#define IP_VS_CONN_F_DROUTE	  0x00400000	/* direct routing */
-                                                /* masquerading otherwise */
 
 #ifdef __KERNEL__
 
@@ -112,7 +106,7 @@ struct ip_vs_rule_user {
 #include <asm/atomic.h>                 /* for atomic_t */
 
 
-#define CONFIG_IP_VS_DEBUG
+/*  #define CONFIG_IP_VS_DEBUG */
 
 #ifdef CONFIG_IP_VS_DEBUG
 extern int ip_vs_get_debug_level(void);
@@ -131,13 +125,21 @@ extern int ip_vs_get_debug_level(void);
 	printk(KERN_WARNING "IPVS: " ## msg)
 
 #ifdef CONFIG_IP_VS_DEBUG
-#define EnterFunction   printk("Enter: %s, %s line %i\n",    \
-                               __FUNCTION__, __FILE__, __LINE__)
-#define LeaveFunction   printk("Leave: %s, %s line %i\n",    \
-                               __FUNCTION__, __FILE__, __LINE__)
+#define EnterFunction(level)                                            \
+    do {                                                                \
+            if (level <= ip_vs_get_debug_level())                       \
+                    printk(KERN_DEBUG "Enter: %s, %s line %i\n",        \
+                           __FUNCTION__, __FILE__, __LINE__);           \
+    } while (0)
+#define LeaveFunction(level)                                            \
+    do {                                                                \
+            if (level <= ip_vs_get_debug_level())                       \
+                        printk(KERN_DEBUG "Leave: %s, %s line %i\n",    \
+                               __FUNCTION__, __FILE__, __LINE__);       \
+    } while (0)
 #else
-#define EnterFunction   do {} while (0)
-#define LeaveFunction   do {} while (0)
+#define EnterFunction(level)   do {} while (0)
+#define LeaveFunction(level)   do {} while (0)
 #endif
 
 /*
@@ -189,8 +191,6 @@ enum {
 
 struct ip_vs_dest;
 struct ip_vs_scheduler;
-struct ip_vs_timeout_table;
-
 
 struct ip_vs_timeout_table {
 	atomic_t refcnt;
@@ -206,6 +206,20 @@ union ip_vs_tphdr {
         __u16 *portp;
 };
 
+
+/*
+ *	Delta sequence info structure
+ *	Each ip_vs_conn has 2 (output AND input seq. changes).
+ *      Only used in the VS/NAT.
+ */
+struct ip_vs_seq {
+        __u32           init_seq;       /* Add delta from this seq */
+        short           delta;          /* Delta in sequence numbers */
+        short           previous_delta; /* Delta in sequence numbers
+                                           before last resized pkt */
+};
+
+
 /*
  *	IP_VS structure allocated for each dynamically scheduled connection
  */
@@ -217,19 +231,21 @@ struct ip_vs_conn {
 	__u16                   protocol;       /* Which protocol (TCP/UDP) */
         __u32                   caddr;          /* client address & port */
         __u16                   cport;
-        __u32                   vaddr;          /* vs address & port */
+        __u32                   vaddr;          /* VS address & port */
         __u16                   vport;
         __u32                   daddr;          /* dest address & port */
         __u16                   dport;
-	struct ip_vs_app        *app;           /* bound IP_VS_app object */
+	struct ip_vs_app        *app;           /* bound ip_vs_app object */
 	void                    *app_data;      /* Application private data */
-	struct ip_vs_conn      *control;       /* Master control connection */
+	struct ip_vs_conn       *control;       /* Master control connection */
 	atomic_t                n_control;      /* Number of controlled ones */
 	unsigned                flags;          /* status flags */
 	unsigned                state;          /* state info */
 	struct ip_vs_timeout_table *timeout_table;
 	struct ip_vs_dest       *dest;          /* real server */
         atomic_t                in_pkts;        /* incoming packet counter */
+        struct ip_vs_seq        in_seq;         /* incoming seq. struct */
+        struct ip_vs_seq        out_seq;        /* outgoing seq. struct */
 };
 
 
@@ -252,7 +268,7 @@ struct ip_vs_service {
 	struct ip_vs_scheduler	*scheduler;    /* bound scheduler object */
 	rwlock_t		sched_lock;    /* lock sched_data */
 	void			*sched_data;   /* scheduler application data */
-	atomic_t		packethandled;	/* packet handled */
+	atomic_t		packethandled; /* packet handled */
 };
 
 
@@ -301,20 +317,28 @@ struct ip_vs_scheduler {
 };
 
 
+/*
+ *	The application module object
+ */
 struct ip_vs_app
 {
-        struct ip_vs_app *next;
-	char *name;		/* name of application proxy */
-        unsigned type;          /* type = proto<<16 | port (host byte order)*/
-        int n_attach;
-        int (*vs_init_1)        /* ip_vs_app initializer */
-                (struct ip_vs_app *, struct ip_vs_conn *);
-        int (*vs_done_1)        /* ip_vs_app fin. */
-                (struct ip_vs_app *, struct ip_vs_conn *);
-        int (*pkt_out)          /* output hook */
-                (struct ip_vs_app *, struct ip_vs_conn *, struct sk_buff **, __u32);
-        int (*pkt_in)           /* input hook */
-                (struct ip_vs_app *, struct ip_vs_conn *, struct sk_buff **, __u32);
+        struct list_head        n_list;   /* d-linked list head */
+        char                    *name;    /* name of application module */
+        unsigned                type;     /* type = proto<<16 | port
+                                             (host byte order)*/
+        /*  atomic_t                refcnt; */   /* reference counter */
+	struct module		*module;  /* THIS_MODULE/NULL */
+        
+        /* ip_vs_app initializer */
+        int (*init_conn)(struct ip_vs_app *, struct ip_vs_conn *);
+        /* ip_vs_app finish */
+        int (*done_conn)(struct ip_vs_app *, struct ip_vs_conn *);
+        /* output hook */
+        int (*pkt_out)(struct ip_vs_app *,
+                       struct ip_vs_conn *, struct sk_buff **);
+        /* input hook */
+        int (*pkt_in)(struct ip_vs_app *,
+                      struct ip_vs_conn *, struct sk_buff **);
 };
 
 
@@ -323,8 +347,8 @@ struct ip_vs_app
  *      (from ip_vs_core.c)
  */
 extern int ip_vs_leave(struct ip_vs_service *svc, struct sk_buff *skb);
-extern int ip_vs_tunnel_xmit(struct sk_buff *skb, __u32 daddr);
-extern int ip_vs_dr_xmit(struct sk_buff *skb, __u32 daddr);
+extern int ip_vs_tunnel_xmit(struct sk_buff *skb, struct ip_vs_conn *cp);
+extern int ip_vs_dr_xmit(struct sk_buff *skb, struct ip_vs_conn *cp);
 extern struct ip_vs_conn * ip_vs_schedule(struct ip_vs_service *svc,
                                           struct iphdr *iph);
 const char *vs_proto_name(unsigned proto);
@@ -339,16 +363,18 @@ const char *vs_proto_name(unsigned proto);
  *     IPVS connection entry hash table
  */
 #ifndef CONFIG_IP_VS_TAB_BITS
-#define CONFIG_IP_VS_TAB_BITS 12
+#define CONFIG_IP_VS_TAB_BITS   12
 #endif
-#define IP_VS_CONN_TAB_BITS	CONFIG_IP_VS_TAB_BITS
-#define IP_VS_CONN_TAB_SIZE  (1 << IP_VS_CONN_TAB_BITS)
-#define IP_VS_CONN_TAB_MASK  (IP_VS_CONN_TAB_SIZE - 1)
+#define IP_VS_CONN_TAB_BITS     CONFIG_IP_VS_TAB_BITS
+#define IP_VS_CONN_TAB_SIZE     (1 << IP_VS_CONN_TAB_BITS)
+#define IP_VS_CONN_TAB_MASK     (IP_VS_CONN_TAB_SIZE - 1)
 
-#define VS_STATE_INPUT	0
-#define VS_STATE_OUTPUT	4
+#define VS_STATE_INPUT	        0
+#define VS_STATE_OUTPUT	        4
 #define VS_STATE_INPUT_ONLY	8
+
 extern struct ip_vs_timeout_table vs_timeout_table_dos;
+extern atomic_t ip_vs_conn_no_cport_cnt;
 
 extern int ip_vs_conn_hash(struct ip_vs_conn *cp);
 extern int ip_vs_conn_unhash(struct ip_vs_conn *cp);
@@ -363,6 +389,7 @@ ip_vs_conn_new(int proto, __u32 maddr, __u16 mport, __u32 saddr,
 extern const char * ip_vs_state_name(int state);
 extern int vs_set_state(struct ip_vs_conn *cp, int state_off,
                         struct iphdr *iph, void *tp);
+extern int ip_vs_conn_listen(struct ip_vs_conn *cp);
 extern int ip_vs_check_template(struct ip_vs_conn *ct);
 extern void ip_vs_secure_tcp_set(int on);
 extern void ip_vs_set_state(struct ip_vs_conn *cp, int new_state);
@@ -430,6 +457,7 @@ ip_vs_control_add(struct ip_vs_conn *cp, struct ip_vs_conn *ctl_cp)
  *      IPVS application functions
  *      (from ip_vs_app.c)
  */
+#define IP_VS_APP_MAX_PORTS  8
 extern int register_ip_vs_app(struct ip_vs_app *mapp,
                               unsigned short proto, __u16 port);
 extern int unregister_ip_vs_app(struct ip_vs_app *mapp);
@@ -517,13 +545,13 @@ extern void ip_vs_sltimer_cleanup(void);
 /*
  *      ip_vs_fwd_tag returns the forwarding tag of the connection
  */
-#define IP_VS_VS_FWD(cp)  (cp->flags & IP_VS_CONN_F_FWD_MASK)
+#define IP_VS_FWD_METHOD(cp)  (cp->flags & IP_VS_CONN_F_FWD_MASK)
 
 extern __inline__ char ip_vs_fwd_tag(struct ip_vs_conn *cp)
 {
         char fwd = 'M';
 
-        switch (IP_VS_VS_FWD(cp)) {
+        switch (IP_VS_FWD_METHOD(cp)) {
         case IP_VS_CONN_F_LOCALNODE:
                 fwd = 'L'; break;
         case IP_VS_CONN_F_TUNNEL:
