@@ -1,26 +1,31 @@
 /*
- * IPVS         application module
+ * IPVS         Application module
  *
- * Version:     $Id: ip_vs_app.c,v 1.1.1.1 2000/05/17 07:55:57 wensong Exp $
+ * Version:     $Id: ip_vs_app.c,v 1.3 2000/07/06 14:57:09 wensong Exp $
  *
- * Authors:     
+ * Authors:     Wensong Zhang <wensong@linuxvirtualserver.org>
  *
  *              This program is free software; you can redistribute it and/or
  *              modify it under the terms of the GNU General Public License
  *              as published by the Free Software Foundation; either version
  *              2 of the License, or (at your option) any later version.
  *
- * Most code here are taken from ip_masq_app.c in kernel 2.2.
+ * Most code here is taken from ip_masq_app.c in kernel 2.2. The difference
+ * is that ip_vs_app module handles the reverse direction (incoming requests
+ * and outgoing responses). The ip_vs_app modules are only used for VS/NAT.
+ *
+ *		IP_MASQ_APP application masquerading module
+ *
+ * 	$Id: ip_vs_app.c,v 1.3 2000/07/06 14:57:09 wensong Exp $
+ *
+ * Author:	Juan Jose Ciarlante, <jjciarla@raiz.uncu.edu.ar>
+ *
  */
 
-#if 0
-
 #ifdef MODULE
 #define EXPORT_SYMTAB
 #endif
 
-#endif
-
 #include <linux/config.h>
 #include <linux/module.h>
 #include <linux/types.h>
@@ -40,318 +45,272 @@
 #include "ip_vs.h"
 //#include <net/ip_vs.h>
 
-#define IP_VS_APP_TAB_SIZE  16 /* must be power of 2 */
+#define IP_VS_APP_TAB_SIZE  16          /* must be power of 2 */
 
 #define IP_VS_APP_HASH(proto, port) ((port^proto) & (IP_VS_APP_TAB_SIZE-1))
-#define IP_VS_APP_TYPE(proto, port) ( proto<<16 | port )
-#define IP_VS_APP_PORT(type)        ( type & 0xffff )
-#define IP_VS_APP_PROTO(type)       ( (type>>16) & 0x00ff )
+#define IP_VS_APP_TYPE(proto, port) (proto<<16 | port)
+#define IP_VS_APP_PORT(type)        (type & 0xffff)
+#define IP_VS_APP_PROTO(type)       ((type>>16) & 0x00ff)
 
 
-#if 0
 EXPORT_SYMBOL(register_ip_vs_app);
 EXPORT_SYMBOL(unregister_ip_vs_app);
-EXPORT_SYMBOL(ip_vs_skb_replace);
-#endif
-
-
-
-/*
- * FIXME: most of code here needs to be rewritten.
- *        because ip_vs_app module just handle request from outside to inside
- */
-
-
 
 
 /*
- * 	will hold masq app. hashed list heads
+ * 	will hold ipvs app. hashed list heads
  */
+struct list_head ip_vs_app_base[IP_VS_APP_TAB_SIZE];
+
+/* lock for ip_vs_app table */
+rwlock_t __ip_vs_app_lock = RW_LOCK_UNLOCKED;
 
-struct ip_vs_app *ip_vs_app_base[IP_VS_APP_TAB_SIZE];
 
 /*
  * 	ip_vs_app registration routine
  *	port: host byte order.
  */
-
 int register_ip_vs_app(struct ip_vs_app *vapp,
                        unsigned short proto, __u16 port)
 {
-        unsigned long flags;
         unsigned hash;
+
         if (!vapp) {
                 IP_VS_ERR("register_ip_vs_app(): NULL arg\n");
                 return -EINVAL;
         }
+
+        MOD_INC_USE_COUNT;
+        
         vapp->type = IP_VS_APP_TYPE(proto, port);
-        vapp->n_attach = 0;
         hash = IP_VS_APP_HASH(proto, port);
 
-        save_flags(flags);
-        cli();
-        vapp->next = ip_vs_app_base[hash];
-        ip_vs_app_base[hash] = vapp;
-        restore_flags(flags);
+        write_lock_bh(&__ip_vs_app_lock);
+        list_add(&vapp->n_list, &ip_vs_app_base[hash]);
+        write_unlock_bh(&__ip_vs_app_lock);
 
         return 0;
 }
 
-#if 0000
-/*
- * 	ip_vs_app unreg. routine.
- */
 
+/*
+ * 	ip_vs_app unregistration routine.
+ */
 int unregister_ip_vs_app(struct ip_vs_app *vapp)
 {
-        struct ip_vs_app **vapp_p;
-        unsigned hash;
-        unsigned long flags;
         if (!vapp) {
                 IP_VS_ERR("unregister_ip_vs_app(): NULL arg\n");
                 return -EINVAL;
         }
+
+#if 0000
         /*
-         * only allow unregistration if it has no attachments
+         * only allow unregistration if it has no referrer.
          */
-        if (vapp->n_attach)  {
-                IP_VS_ERR("unregister_ip_vs_app(): has %d attachments. failed\n",
-                          vapp->n_attach);
+        if (atomic_read(&vapp->refcnt) > 0)  {
+                IP_VS_ERR("unregister_ip_vs_app(): has %d referrers. failed\n",
+                          atomic_read(&vapp->refcnt));
                 return -EINVAL;
         }
-        hash = IP_VS_APP_HASH(IP_VS_APP_PROTO(vapp->type), IP_MASQ_APP_PORT(vapp->type));
+#endif
+        
+        write_lock_bh(&__ip_vs_app_lock);
+        list_del(&vapp->n_list);
+        write_unlock_bh(&__ip_vs_app_lock);
+        
+        MOD_DEC_USE_COUNT;
 
-        save_flags(flags);
-        cli();
-        for (vapp_p = &ip_vs_app_base[hash]; *vapp_p ; vapp_p = &(*vapp_p)->next)
-                if (vapp == (*vapp_p))  {
-                        *vapp_p = vapp->next;
-                        restore_flags(flags);
-                        return 0;
-                }
-
-        restore_flags(flags);
-        IP_VS_ERR("unregister_ip_vs_app(proto=%s,port=%u): not hashed!\n",
-                  vs_proto_name(IP_VS_APP_PROTO(vapp->type)), IP_VS_APP_PORT(vapp->type));
-        return -EINVAL;
+        return 0;
 }
 
 
 /*
  *	get ip_vs_app object by its proto and port (net byte order).
  */
-
 struct ip_vs_app * ip_vs_app_get(unsigned short proto, __u16 port)
 {
+        struct list_head *e;
         struct ip_vs_app *vapp;
         unsigned hash;
         unsigned type;
 
         port = ntohs(port);
-        type = IP_VS_APP_TYPE(proto,port);
-        hash = IP_VS_APP_HASH(proto,port);
-        for(vapp = ip_vs_app_base[hash]; vapp ; vapp = vapp->next) {
-                if (type == vapp->type) return vapp;
+        type = IP_VS_APP_TYPE(proto, port);
+        hash = IP_VS_APP_HASH(proto, port);
+
+        read_lock_bh(&__ip_vs_app_lock);
+        
+        list_for_each(e, &ip_vs_app_base[hash]) {
+                vapp = list_entry(e, struct ip_vs_app, n_list);
+
+                /*
+                 * Test and MOD_INC_USE_COUNT atomically
+                 */
+                if (!try_inc_mod_count(vapp->module)) {
+                        /*
+                         * This application module is just deleted
+                         */
+                        continue;
+                }
+                if (type == vapp->type) {
+                        read_unlock_bh(&__ip_vs_app_lock);
+                        return vapp;
+                }
+                
+                __MOD_DEC_USE_COUNT(vapp->module);
         }
+
+        read_unlock_bh(&__ip_vs_app_lock);
         return NULL;
 }
 
-/*
- *	ip_vs_app object binding related funcs.
- */
 
 /*
- * 	change ip_vs_app object's number of bindings
+ *	Bind ip_vs_conn to its ip_vs_app based on proto and dport,
+ *	and call the ip_vs_app constructor.
  */
-
-static __inline__ int ip_vs_app_bind_chg(struct ip_vs_app *vapp, int delta)
-{
-        unsigned long flags;
-        int n_at;
-        if (!vapp) return -1;
-        save_flags(flags);
-        cli();
-        n_at = vapp->n_attach + delta;
-        if (n_at < 0) {
-                restore_flags(flags);
-                IP_VS_ERR("ip_vs_app: tried to set n_attach < 0 for (proto=%s,port==%d) ip_vs_app object.\n",
-                          vs_proto_name(IP_VS_APP_PROTO(vapp->type)),
-                          IP_VS_APP_PORT(vapp->type));
-                return -1;
-        }
-        vapp->n_attach = n_at;
-        restore_flags(flags);
-        return 0;
-}
-        
-#endif
-
-/*
- *	Bind ip_vs to its ip_vs_app based on proto and dport ALREADY
- *	set in ip_vs struct. Also calls constructor.
- */
-
 struct ip_vs_app * ip_vs_bind_app(struct ip_vs_conn *cp)
 {
-#if 0000
-        struct ip_vs_app * vapp;
+        struct ip_vs_app *vapp;
 
-	if (cp->protocol != IPPROTO_TCP && cp->protocol != IPPROTO_UDP)
-		return NULL;
+        /* no need to bind app if its forwarding method is not NAT */
+        if (IP_VS_FWD_METHOD(cp) != IP_VS_CONN_F_MASQ)
+                return NULL;
 
-        vapp = ip_vs_app_get(cp->protocol, cp->dport);
+        if (cp->protocol != IPPROTO_TCP && cp->protocol != IPPROTO_UDP)
+                return NULL;
 
-
-#ifdef CONFIG_IP_MASQUERADE_VS
-        if (vapp == NULL)
-                /*
-                 * check for virtual service.
-                 */
-                vapp = ip_vs_app_get(cp->protocol, cp->mport);
-#endif
+        /*
+         *	don't allow binding if already bound
+         */
+        if (cp->app != NULL) {
+                IP_VS_ERR("ip_vs_bind_app(): "
+                          "called for already bound object.\n");
+                return cp->app;
+        }
+        
+        vapp = ip_vs_app_get(cp->protocol, cp->vport);
 
         if (vapp != NULL) {
-                /*
-                 *	don't allow binding if already bound
-                 */
-
-                if (cp->app != NULL) {
-                        IP_VS_ERR("ip_vs_bind_app() called for already bound object.\n");
-                        return cp->app;
-                }
-
+                /*  atomic_inc(&vapp->refcnt); */
                 cp->app = vapp;
-                if (vapp->masq_init_1) vapp->masq_init_1(vapp, cp);
-                ip_vs_app_bind_chg(vapp, +1);
+                
+                if (vapp->init_conn)
+                        vapp->init_conn(vapp, cp);
         }
         return vapp;
-#endif
-        return NULL;
 }
 
+
 /*
  * 	Unbind cp from type object and call cp destructor (does not kfree()).
  */
-
 int ip_vs_unbind_app(struct ip_vs_conn *cp)
 {
-#if 0000
-        struct ip_vs_app * vapp;
-        vapp = cp->app;
+        struct ip_vs_app *vapp = cp->app;
 
-	if (cp->protocol != IPPROTO_TCP && cp->protocol != IPPROTO_UDP)
-		return 0;
+        if (cp->protocol != IPPROTO_TCP && cp->protocol != IPPROTO_UDP)
+                return 0;
 
         if (vapp != NULL) {
-                if (vapp->masq_done_1) vapp->masq_done_1(vapp, cp);
+                if (vapp->done_conn)
+                        vapp->done_conn(vapp, cp);
                 cp->app = NULL;
-                ip_vs_app_bind_chg(vapp, -1);
+                /*  atomic_dec(&vapp->refcnt); */
+                __MOD_DEC_USE_COUNT(vapp->module);
         }
         return (vapp != NULL);
-#endif
-        return 0;
 }
 
 
-#if 0000
 /*
  *	Fixes th->seq based on ip_vs_seq info.
  */
-static __inline__ void masq_fix_seq(const struct ip_vs_seq *cp_seq, struct tcphdr *th)
+static inline void vs_fix_seq(const struct ip_vs_seq *vseq, struct tcphdr *th)
 {
-        __u32 seq;
+        __u32 seq = ntohl(th->seq);
 
-        seq = ntohl(th->seq);
-
-	/*
-	 * 	Adjust seq with delta-offset for all packets after
+        /*
+         * 	Adjust seq with delta-offset for all packets after
          * 	the most recent resized pkt seq and with previous_delta offset
          *	for all packets	before most recent resized pkt seq.
-	 */
-	if (cp_seq->delta || cp_seq->previous_delta) {
-		if(after(seq,cp_seq->init_seq) ) {
-			th->seq = htonl(seq + cp_seq->delta);
-			IP_VS_DEBUG(1, "masq_fix_seq() : added delta (%d) to seq\n",cp_seq->delta);
-		} else {
-			th->seq = htonl(seq + cp_seq->previous_delta);
-			IP_VS_DEBUG(1, "masq_fix_seq() : added previous_delta (%d) to seq\n",cp_seq->previous_delta);
-		}
-	}
-
-
+         */
+        if (vseq->delta || vseq->previous_delta) {
+                if(after(seq, vseq->init_seq)) {
+                        th->seq = htonl(seq + vseq->delta);
+                        IP_VS_DBG(1, "vs_fix_seq(): added delta (%d) to seq\n",
+                                  vseq->delta);
+                } else {
+                        th->seq = htonl(seq + vseq->previous_delta);
+                        IP_VS_DBG(1, "vs_fix_seq(): added previous_delta "
+                                  "(%d) to seq\n", vseq->previous_delta);
+                }
+        }
 }
 
+
 /*
  *	Fixes th->ack_seq based on ip_vs_seq info.
  */
-
-static __inline__ void masq_fix_ack_seq(const struct ip_vs_seq *cp_seq, struct tcphdr *th)
+static inline void
+vs_fix_ack_seq(const struct ip_vs_seq *vseq, struct tcphdr *th)
 {
-        __u32 ack_seq;
-
-        ack_seq=ntohl(th->ack_seq);
+        __u32 ack_seq = ntohl(th->ack_seq);
 
         /*
          * Adjust ack_seq with delta-offset for
          * the packets AFTER most recent resized pkt has caused a shift
          * for packets before most recent resized pkt, use previous_delta
          */
-
-        if (cp_seq->delta || cp_seq->previous_delta) {
-                if(after(ack_seq,cp_seq->init_seq)) {
-                        th->ack_seq = htonl(ack_seq-cp_seq->delta);
-                        IP_VS_DEBUG(1, "masq_fix_ack_seq() : subtracted delta (%d) from ack_seq\n",cp_seq->delta);
+        if (vseq->delta || vseq->previous_delta) {
+                if(after(ack_seq, vseq->init_seq)) {
+                        th->ack_seq = htonl(ack_seq - vseq->delta);
+                        IP_VS_DBG(1, "vs_fix_ack_seq(): subtracted delta "
+                                  "(%d) from ack_seq\n", vseq->delta);
 
                 } else {
-                        th->ack_seq = htonl(ack_seq-cp_seq->previous_delta);
-                        IP_VS_DEBUG(1, "masq_fix_ack_seq() : subtracted previous_delta (%d) from ack_seq\n",cp_seq->previous_delta);
+                        th->ack_seq = htonl(ack_seq - vseq->previous_delta);
+                        IP_VS_DBG(1, "vs_fix_ack_seq(): subtracted "
+                                  "previous_delta (%d) from ack_seq\n",
+                                  vseq->previous_delta);
                 }
         }
-
 }
 
+
 /*
  *	Updates ip_vs_seq if pkt has been resized
  *	Assumes already checked proto==IPPROTO_TCP and diff!=0.
  */
-
-static __inline__ void masq_seq_update(struct ip_vs_conn *cp, struct ip_vs_seq *cp_seq, unsigned mflag, __u32 seq, int diff)
+static inline void vs_seq_update(struct ip_vs_conn *cp, struct ip_vs_seq *vseq,
+                                 unsigned flag, __u32 seq, int diff)
 {
-        /* if (diff == 0) return; */
-
-        if ( !(cp->flags & mflag) || after(seq, cp_seq->init_seq))
-        {
-                cp_seq->previous_delta=cp_seq->delta;
-                cp_seq->delta+=diff;
-                cp_seq->init_seq=seq;
-                cp->flags |= mflag;
+        if ( !(cp->flags & flag) || after(seq, vseq->init_seq)) {
+                vseq->previous_delta = vseq->delta;
+                vseq->delta += diff;
+                vseq->init_seq = seq;
+                cp->flags |= flag;
         }
 }
 
-#endif
 
 /*
  *	Output pkt hook. Will call bound ip_vs_app specific function
- *	called by ip_fw_masquerade(), assumes previously checked cp!=NULL
+ *	called by ip_vs_out(), assumes previously checked cp!=NULL
  *	returns (new - old) skb->len diff.
  */
-
 int ip_vs_app_pkt_out(struct ip_vs_conn *cp, struct sk_buff **skb_p)
 {
-#if 0000
         struct ip_vs_app * vapp;
-        struct iphdr *iph;
-	struct tcphdr *th;
         int diff;
+        struct iphdr *iph;
+        struct tcphdr *th;
         __u32 seq;
 
         /*
-         *	check if application masquerading is bound to
-         *	this ip_vs.
-         *	assumes that once an ip_vs is bound,
-         *	it will not be unbound during its life.
+         *	check if application module is bound to
+         *	this ip_vs_conn.
          */
-
         if ( (vapp = cp->app) == NULL)
                 return 0;
 
@@ -361,63 +320,54 @@ int ip_vs_app_pkt_out(struct ip_vs_conn *cp, struct sk_buff **skb_p)
         /*
          *	Remember seq number in case this pkt gets resized
          */
-
         seq = ntohl(th->seq);
 
         /*
          *	Fix seq stuff if flagged as so.
          */
-
         if (cp->protocol == IPPROTO_TCP) {
-                if (cp->flags & IP_VS_F_OUT_SEQ)
-                        masq_fix_seq(&cp->out_seq, th);
-                if (cp->flags & IP_VS_F_IN_SEQ)
-                        masq_fix_ack_seq(&cp->in_seq, th);
+                if (cp->flags & IP_VS_CONN_F_OUT_SEQ)
+                        vs_fix_seq(&cp->out_seq, th);
+                if (cp->flags & IP_VS_CONN_F_IN_SEQ)
+                        vs_fix_ack_seq(&cp->in_seq, th);
         }
-
+        
         /*
          *	Call private output hook function
          */
-
         if ( vapp->pkt_out == NULL )
                 return 0;
 
-        diff = vapp->pkt_out(vapp, cp, skb_p, maddr);
+        diff = vapp->pkt_out(vapp, cp, skb_p);
 
         /*
          *	Update ip_vs seq stuff if len has changed.
          */
-
         if (diff != 0 && cp->protocol == IPPROTO_TCP)
-                masq_seq_update(cp, &cp->out_seq, IP_VS_F_OUT_SEQ, seq, diff);
+                vs_seq_update(cp, &cp->out_seq,
+                              IP_VS_CONN_F_OUT_SEQ, seq, diff);
 
         return diff;
-#endif
-        return 0;
 }
 
+
 /*
  *	Input pkt hook. Will call bound ip_vs_app specific function
  *	called by ip_fw_demasquerade(), assumes previously checked cp!=NULL.
  *	returns (new - old) skb->len diff.
  */
-
 int ip_vs_app_pkt_in(struct ip_vs_conn *cp, struct sk_buff **skb_p)
 {
-#if 0000
         struct ip_vs_app * vapp;
-        struct iphdr *iph;
-	struct tcphdr *th;
         int diff;
+        struct iphdr *iph;
+        struct tcphdr *th;
         __u32 seq;
 
         /*
-         *	check if application masquerading is bound to
-         *	this ip_vs.
-         *	assumes that once an ip_vs is bound,
-         *	it will not be unbound during its life.
+         *	check if application module is bound to
+         *	this ip_vs_conn.
          */
-
         if ( (vapp = cp->app) == NULL)
                 return 0;
 
@@ -427,87 +377,83 @@ int ip_vs_app_pkt_in(struct ip_vs_conn *cp, struct sk_buff **skb_p)
         /*
          *	Remember seq number in case this pkt gets resized
          */
-
         seq = ntohl(th->seq);
 
         /*
          *	Fix seq stuff if flagged as so.
          */
-
         if (cp->protocol == IPPROTO_TCP) {
-                if (cp->flags & IP_VS_F_IN_SEQ)
-                        masq_fix_seq(&cp->in_seq, th);
-                if (cp->flags & IP_VS_F_OUT_SEQ)
-                        masq_fix_ack_seq(&cp->out_seq, th);
+                if (cp->flags & IP_VS_CONN_F_IN_SEQ)
+                        vs_fix_seq(&cp->in_seq, th);
+                if (cp->flags & IP_VS_CONN_F_OUT_SEQ)
+                        vs_fix_ack_seq(&cp->out_seq, th);
         }
-
+        
         /*
          *	Call private input hook function
          */
-
         if ( vapp->pkt_in == NULL )
                 return 0;
 
-        diff = vapp->pkt_in(vapp, cp, skb_p, maddr);
+        diff = vapp->pkt_in(vapp, cp, skb_p);
 
         /*
          *	Update ip_vs seq stuff if len has changed.
          */
-
         if (diff != 0 && cp->protocol == IPPROTO_TCP)
-                masq_seq_update(cp, &cp->in_seq, IP_VS_F_IN_SEQ, seq, diff);
-
+                vs_seq_update(cp, &cp->in_seq,
+                              IP_VS_CONN_F_IN_SEQ, seq, diff);
+        
         return diff;
-#endif
-        return 0;
 }
 
-/*
- *	/proc/ip_vs_app entry function
- */
 
+/*
+ *	/proc/net/ip_vs_app entry function
+ */
 int ip_vs_app_getinfo(char *buffer, char **start, off_t offset, int length)
 {
-        off_t pos=0, begin=0;
+        off_t pos=0, begin;
         int len=0;
-/*          struct ip_vs_app * vapp; */
-/*          unsigned idx; */
+        struct ip_vs_app * vapp;
+        int idx, size;
+        struct list_head *e;
 
-	if (offset < 40)
-		len=sprintf(buffer, "%-39s\n", "prot port    n_attach name");
-	pos = 40;
+        size = sprintf(buffer, "prot port    usecnt name\n");
+        pos += size;
+        len += size;
 
-#if 0000
-        for (idx=0 ; idx < IP_VS_APP_TAB_SIZE; idx++)
-                for (vapp = ip_vs_app_base[idx]; vapp ; vapp = vapp->next) {
-			/*
-			 * If you change the length of this sprintf, then all
-			 * the length calculations need fixing too!
-			 * Line length = 40 (3 + 2 + 7 + 1 + 7 + 1 + 2 + 17)
-			 */
-			pos += 40;
-			if (pos < offset)
-				continue;
+        for (idx=0 ; idx < IP_VS_APP_TAB_SIZE; idx++) {
+                read_lock_bh(&__ip_vs_app_lock);
+                list_for_each (e, &ip_vs_app_base[idx]) {
+                        vapp = list_entry(e, struct ip_vs_app, n_list);
 
-                        len += sprintf(buffer+len, "%-3s  %-7u %-7d  %-17s\n",
-                                       vs_proto_name(IP_VS_APP_PROTO(vapp->type)),
-                                       IP_VS_APP_PORT(vapp->type), vapp->n_attach,
-				       vapp->name);
+                        size += sprintf(buffer+len, "%-3s  %-7u %-6d %-17s\n",
+                                        vs_proto_name(
+                                                IP_VS_APP_PROTO(vapp->type)),
+                                        IP_VS_APP_PORT(vapp->type),
+                                        GET_USE_COUNT(vapp->module),
+                                        vapp->name);
 
-                        if(len >= length)
+                        len += size;
+                        pos += size;
+                        if (pos <= offset)
+                                len=0;
+                        if (pos >= offset+length) {
+                                read_unlock_bh(&__ip_vs_app_lock);
                                 goto done;
+                        }
                 }
+                read_unlock_bh(&__ip_vs_app_lock);
+        }
+
   done:
-
-
-#endif
-        
-        begin = len - (pos - offset);
-        *start = buffer + begin;
-        len -= begin;
-        if (len > length)
-                len = length;
-        return len;
+	begin = len - (pos - offset);
+	*start = buffer + begin;
+	len -= begin;
+	if(len>length)
+		len = length;
+	return len;
 }
 
 
@@ -517,40 +463,38 @@ int ip_vs_app_getinfo(char *buffer, char **start, off_t offset, int length)
  *	       be done if n_len < o_len, unless some extra space
  *	       were already allocated at driver level :P .
  */
-
-static struct sk_buff * skb_replace(struct sk_buff *skb, int pri, char *o_buf, int o_len, char *n_buf, int n_len)
+static struct sk_buff *skb_replace(struct sk_buff *skb, int pri, char *o_buf,
+                                   int o_len, char *n_buf, int n_len)
 {
         int maxsize, diff, o_offset;
         struct sk_buff *n_skb;
-	int offset;
-
-	maxsize = skb->truesize;
+        int offset;
 
+        maxsize = skb->truesize;
         diff = n_len - o_len;
         o_offset = o_buf - (char*) skb->data;
 
-	if (maxsize <= n_len) {
-	    if (diff != 0) {
-		memcpy(skb->data + o_offset + n_len,o_buf + o_len,
-		       skb->len - (o_offset + o_len));
-	    }
+        if (maxsize <= n_len) {
+                if (diff != 0) {
+                        memcpy(skb->data + o_offset + n_len,o_buf + o_len,
+                               skb->len - (o_offset + o_len));
+                }
 
-	    memcpy(skb->data + o_offset, n_buf, n_len);
+                memcpy(skb->data + o_offset, n_buf, n_len);
 
-	    n_skb    = skb;
-	    skb->len = n_len;
-	    skb->end = skb->head+n_len;
-	} else {
+                n_skb    = skb;
+                skb->len = n_len;
+                skb->end = skb->head+n_len;
+        } else {
                 /*
                  * 	Sizes differ, make a copy.
                  *
                  *	FIXME: move this to core/sbuff.c:skb_grow()
                  */
-
                 n_skb = alloc_skb(MAX_HEADER + skb->len + diff, pri);
                 if (n_skb == NULL) {
                         IP_VS_ERR("skb_replace(): no room left (from %p)\n",
-                               __builtin_return_address(0));
+                                  __builtin_return_address(0));
                         return skb;
 
                 }
@@ -570,7 +514,7 @@ static struct sk_buff * skb_replace(struct sk_buff *skb, int pri, char *o_buf, i
                 n_skb->pkt_type = skb->pkt_type;
                 n_skb->protocol = skb->protocol;
                 n_skb->ip_summed = skb->ip_summed;
-		n_skb->dst = dst_clone(skb->dst);
+                n_skb->dst = dst_clone(skb->dst);
 
                 /*
                  * Copy pkt in new buffer
@@ -586,6 +530,7 @@ static struct sk_buff * skb_replace(struct sk_buff *skb, int pri, char *o_buf, i
                  */
                 kfree_skb(skb);
         }
+
         return n_skb;
 }
 
@@ -593,12 +538,15 @@ static struct sk_buff * skb_replace(struct sk_buff *skb, int pri, char *o_buf, i
 /*
  *	calls skb_replace() and update ip header if new skb was allocated
  */
-struct sk_buff * ip_vs_skb_replace(struct sk_buff *skb, int pri, char *o_buf, int o_len, char *n_buf, int n_len)
+struct sk_buff *ip_vs_skb_replace(struct sk_buff *skb, int pri, char *o_buf,
+                                  int o_len, char *n_buf, int n_len)
 {
         int diff;
         struct sk_buff *n_skb;
         unsigned skb_len;
 
+	EnterFunction(1);
+
         diff = n_len - o_len;
         n_skb = skb_replace(skb, pri, o_buf, o_len, n_buf, n_len);
         skb_len = skb->len;
@@ -616,14 +564,22 @@ struct sk_buff * ip_vs_skb_replace(struct sk_buff *skb, int pri, char *o_buf, in
                 iph->check = ip_fast_csum((unsigned char *)iph, iph->ihl);
                 iph->tot_len = htons(skb_len + diff);
         }
+        
+        LeaveFunction(1);
         return n_skb;
 }
 
 
 int ip_vs_app_init(void)
 {
+        int idx;
+
+        for (idx=0 ; idx < IP_VS_APP_TAB_SIZE; idx++) {
+                INIT_LIST_HEAD(&ip_vs_app_base[idx]);
+        }
+        
         /* we will replace it with proc_net_ipvs_create() soon */
-	proc_net_create("ip_vs_app", 0, ip_vs_app_getinfo);
+        proc_net_create("ip_vs_app", 0, ip_vs_app_getinfo);
         return 0;
 }
 
