@@ -5,7 +5,7 @@
  *              high-performance and highly available server based on a
  *              cluster of servers.
  *
- * Version:     $Id: ip_vs_conn.c,v 1.1.1.1 2000/05/19 07:46:22 wensong Exp $
+ * Version:     $Id: ip_vs_conn.c,v 1.3 2000/07/06 14:57:09 wensong Exp $
  *
  * Authors:     Wensong Zhang <wensong@iinchina.net>
  *              Peter Kese <peter.kese@ijs.si>
@@ -18,7 +18,7 @@
  *
  * The IPVS code for kernel 2.2 was done by Wensong Zhang and Peter Kese,
  * with changes/fixes from Julian Anastasov, Lars Marowsky-Bree, Horms
- * and others. Many codes here are taken from IP MASQ code of kernel 2.2.
+ * and others. Many code here is taken from IP MASQ code of kernel 2.2.
  *
  * Changes:
  *
@@ -34,17 +34,22 @@
 #include <linux/tcp.h>                  /* for tcphdr */
 #include <linux/in.h>
 #include <linux/proc_fs.h>              /* for proc_net_* */
-#include <linux/netdevice.h>            /* for local_bh_disable/local_bh_enable, I don't know how to include header file elegently now */
+#include <asm/softirq.h>                /* for local_bh_* */
 
 #include "ip_vs.h"
 //#include <net/ip_vs.h>
 
 
 /*
- *  Hash table: for input and output packets lookups of IPVS
+ *  Connection hash table: for input and output packets lookups of IPVS
  */
 struct list_head *ip_vs_table;
 
+/*
+ *  No client port connection couter
+ */
+atomic_t ip_vs_conn_no_cport_cnt = ATOMIC_INIT(0);
+
 
 /*
  * 	Set ip_vs_conn expiration (deletion) and adds timer,
@@ -193,11 +198,11 @@ int ip_vs_conn_unhash(struct ip_vs_conn *cp)
  *	d_addr, d_port: pkt dest address (load balancer)
  *  Caller must lock tables
  */
-struct ip_vs_conn *ip_vs_conn_in_get
+static inline struct ip_vs_conn *__ip_vs_conn_in_get
 (int protocol, __u32 s_addr, __u16 s_port, __u32 d_addr, __u16 d_port)
 {
         unsigned hash;
-        struct ip_vs_conn *cp=NULL;
+        struct ip_vs_conn *cp;
         struct list_head *l,*e;
 
         hash = ip_vs_conn_hashkey(protocol, s_addr, s_port);
@@ -213,21 +218,32 @@ struct ip_vs_conn *ip_vs_conn_in_get
                         /* HIT */
                         atomic_inc(&cp->refcnt);
                         __ip_vs_set_expire(cp, 0);
-                        break;
+                        ct_read_unlock(hash);
+                        return cp;
                 }
         }
 
         ct_read_unlock(hash);
-        
-        IP_VS_DBG(2, "lookup/in %d %u.%u.%u.%u:%d->%u.%u.%u.%u:%d %s\n",
-                  protocol,
-                  NIPQUAD(s_addr),
-                  ntohs(s_port),
-                  NIPQUAD(d_addr),
-                  ntohs(d_port),
-                  cp?"hit":"not hit");
 
-        return cp;
+        return NULL;
+}
+
+struct ip_vs_conn *ip_vs_conn_in_get
+(int protocol, __u32 s_addr, __u16 s_port, __u32 d_addr, __u16 d_port)
+{
+        struct ip_vs_conn *ret;
+
+        ret = __ip_vs_conn_in_get(protocol, s_addr, s_port, d_addr, d_port);
+        if (!ret && atomic_read(&ip_vs_conn_no_cport_cnt))
+                ret = __ip_vs_conn_in_get(protocol, s_addr, 0, d_addr, d_port);
+        
+        IP_VS_DBG(2, "lookup/in %s %u.%u.%u.%u:%d->%u.%u.%u.%u:%d %s\n",
+                  vs_proto_name(protocol),
+                  NIPQUAD(s_addr), ntohs(s_port),
+                  NIPQUAD(d_addr), ntohs(d_port),
+                  ret?"hit":"not hit");
+
+        return ret;
 }
 
 
@@ -242,7 +258,7 @@ struct ip_vs_conn *ip_vs_conn_out_get
 (int protocol, __u32 s_addr, __u16 s_port, __u32 d_addr, __u16 d_port)
 {
         unsigned hash;
-        struct ip_vs_conn *cp=NULL;
+        struct ip_vs_conn *cp, *ret=NULL;
         struct list_head *l,*e;
 
         /*	
@@ -261,21 +277,20 @@ struct ip_vs_conn *ip_vs_conn_out_get
                         /* HIT */
                         atomic_inc(&cp->refcnt);
                         __ip_vs_set_expire(cp, 0);
+                        ret = cp;
                         break;
                 }
         }
 
         ct_read_unlock(hash);
         
-        IP_VS_DBG(2, "lookup/out %d %u.%u.%u.%u:%d->%u.%u.%u.%u:%d %s\n",
-                  protocol,
-                  NIPQUAD(s_addr),
-                  ntohs(s_port),
-                  NIPQUAD(d_addr),
-                  ntohs(d_port),
-                  cp?"hit":"not hit");
+        IP_VS_DBG(2, "lookup/out %s %u.%u.%u.%u:%d->%u.%u.%u.%u:%d %s\n",
+                  vs_proto_name(protocol),
+                  NIPQUAD(s_addr), ntohs(s_port),
+                  NIPQUAD(d_addr), ntohs(d_port),
+                  ret?"hit":"not hit");
 
-        return cp;
+        return ret;
 }
 
 
@@ -471,7 +486,7 @@ static inline int vs_tcp_state_idx(struct tcphdr *th, int state_off)
 }
 
 
-static int vs_set_state_timeout(struct ip_vs_conn *cp, int state)
+static inline int vs_set_state_timeout(struct ip_vs_conn *cp, int state)
 {
 	struct ip_vs_timeout_table *vstim = cp->timeout_table;
 	int scale;
@@ -494,7 +509,7 @@ static int vs_set_state_timeout(struct ip_vs_conn *cp, int state)
 }
 
 
-static int
+static inline int
 vs_tcp_state(struct ip_vs_conn *cp, int state_off, struct tcphdr *th)
 {
 	int state_idx;
@@ -524,7 +539,7 @@ vs_tcp_state(struct ip_vs_conn *cp, int state_off, struct tcphdr *th)
                 struct ip_vs_dest *dest = cp->dest;
 
                 IP_VS_DBG(1, "%s %s [%c%c%c%c] %u.%u.%u.%u:%d->"
-                          "%u.%u.%u.%u:%d state: %s->%s\n",
+                          "%u.%u.%u.%u:%d state: %s->%s cnt:%d\n",
                           vs_proto_name(cp->protocol),
                           (state_off==VS_STATE_OUTPUT)?"output ":"input ",
                           th->syn? 'S' : '.',
@@ -534,7 +549,8 @@ vs_tcp_state(struct ip_vs_conn *cp, int state_off, struct tcphdr *th)
                           NIPQUAD(cp->daddr), ntohs(cp->dport),
                           NIPQUAD(cp->caddr), ntohs(cp->cport),
                           ip_vs_state_name(cp->state),
-                          ip_vs_state_name(new_state));
+                          ip_vs_state_name(new_state),
+                          atomic_read(&cp->refcnt));
                 if (dest) {
                         if (!(cp->flags & IP_VS_CONN_F_INACTIVE) &&
                             (new_state != IP_VS_S_ESTABLISHED)) {
@@ -547,15 +563,6 @@ vs_tcp_state(struct ip_vs_conn *cp, int state_off, struct tcphdr *th)
                                 atomic_dec(&dest->inactconns);
                                 cp->flags &= ~IP_VS_CONN_F_INACTIVE;
                         }
-        
-                        IP_VS_DBG(1, "Set-state fwd:%c s:%s c:%08X:%04X "
-                                  "v:%08X:%04X d:%08X:%04X flg:%X cnt:%d\n",
-                                  ip_vs_fwd_tag(cp),
-                                  ip_vs_state_name(cp->state),
-                                  ntohl(cp->caddr),ntohs(cp->cport),
-                                  ntohl(cp->vaddr),ntohs(cp->vport),
-                                  ntohl(cp->daddr),ntohs(cp->dport),
-                                  cp->flags, atomic_read(&cp->refcnt));
                 }
 	}
 
@@ -581,21 +588,22 @@ int vs_set_state(struct ip_vs_conn *cp,
 }
 
 
+/*
+ *	Set LISTEN timeout. (ip_vs_conn_put will setup timer)
+ */
+int ip_vs_conn_listen(struct ip_vs_conn *cp)
+{
+	vs_set_state_timeout(cp, IP_VS_S_LISTEN);
+	return cp->timeout;
+}
+
+
 /*
  *  Bind a connection entry with a virtual service destination
  *  Called when a new connection entry is created for VS.
  */
 void ip_vs_bind_dest(struct ip_vs_conn *cp, struct ip_vs_dest *dest)
 {
-        IP_VS_DBG(1, "Bind-dest fwd:%c s:%s c:%08X:%04X v:%08X:%04X "
-                  "d:%08X:%04X flg:%X cnt:%d destcnt:%d\n",
-                  ip_vs_fwd_tag(cp), ip_vs_state_name(cp->state),
-                  ntohl(cp->caddr),ntohs(cp->cport),
-                  ntohl(cp->vaddr),ntohs(cp->vport),
-                  ntohl(cp->daddr),ntohs(cp->dport),
-                  cp->flags, atomic_read(&cp->refcnt),
-                  atomic_read(&dest->refcnt));
-
         cp->flags |= atomic_read(&dest->conn_flags);
         cp->dest = dest;
 
@@ -603,6 +611,16 @@ void ip_vs_bind_dest(struct ip_vs_conn *cp, struct ip_vs_dest *dest)
          *    Increase the refcnt counter of the dest.
          */
         atomic_inc(&dest->refcnt);
+
+        IP_VS_DBG(1, "Bind-dest %s c:%u.%u.%u.%u:%d v:%u.%u.%u.%u:%d "
+                  "d:%u.%u.%u.%u:%d fwd:%c s:%s flg:%X cnt:%d destcnt:%d\n",
+                  vs_proto_name(cp->protocol),
+                  NIPQUAD(cp->caddr), ntohs(cp->cport),
+                  NIPQUAD(cp->vaddr), ntohs(cp->vport),
+                  NIPQUAD(cp->daddr), ntohs(cp->dport),
+                  ip_vs_fwd_tag(cp), ip_vs_state_name(cp->state),
+                  cp->flags, atomic_read(&cp->refcnt),
+                  atomic_read(&dest->refcnt));
 }
 
 
@@ -614,12 +632,13 @@ void ip_vs_unbind_dest(struct ip_vs_conn *cp)
 {
         struct ip_vs_dest *dest = cp->dest;
         
-        IP_VS_DBG(1, "Unbind-dest fwd:%c s:%s c:%08X:%04X v:%08X:%04X "
-                  "d:%08X:%04X flg:%X cnt:%d destcnt:%d\n",
+        IP_VS_DBG(1, "Unbind-dest %s c:%u.%u.%u.%u:%d v:%u.%u.%u.%u:%d "
+                  "d:%u.%u.%u.%u:%d fwd:%c s:%s flg:%X cnt:%d destcnt:%d\n",
+                  vs_proto_name(cp->protocol),
+                  NIPQUAD(cp->caddr), ntohs(cp->cport),
+                  NIPQUAD(cp->vaddr), ntohs(cp->vport),
+                  NIPQUAD(cp->daddr), ntohs(cp->dport),
                   ip_vs_fwd_tag(cp), ip_vs_state_name(cp->state),
-                  ntohl(cp->caddr),ntohs(cp->cport),
-                  ntohl(cp->vaddr),ntohs(cp->vport),
-                  ntohl(cp->daddr),ntohs(cp->dport),
                   cp->flags, atomic_read(&cp->refcnt),
                   atomic_read(&dest->refcnt));
 
@@ -660,13 +679,13 @@ int ip_vs_check_template(struct ip_vs_conn *ct)
          */
         if ((dest == NULL) ||
             !(dest->flags & IP_VS_DEST_F_AVAILABLE)) {
-                IP_VS_DBG(1, "check_template: dest not available for prot %d "
-                          "src %08X:%04X dest %08X:%04X -> %08X:%04X\n",
-                          ct->protocol,
-                          ntohl(ct->caddr), ntohs(ct->cport),
-                          ntohl(ct->vaddr), ntohs(ct->vport),
-                          (dest!=NULL)? ntohl(dest->addr):0,
-                          (dest!=NULL)? ntohs(dest->port):0);
+                IP_VS_DBG(1, "check_template: dest not available for "
+                          "protocol %s s:%u.%u.%u.%u:%d v:%u.%u.%u.%u:%d "
+                          "-> d:%u.%u.%u.%u:%d\n",
+                          vs_proto_name(ct->protocol),
+                          NIPQUAD(ct->caddr), ntohs(ct->cport),
+                          NIPQUAD(ct->vaddr), ntohs(ct->vport),
+                          NIPQUAD(ct->daddr), ntohs(ct->dport));
 
                 /*
                  * Invalidate the connection template
@@ -767,11 +786,10 @@ ip_vs_conn_new(int proto, __u32 caddr, __u16 cport, __u32 vaddr, __u16 vport,
                __u32 daddr, __u16 dport, unsigned flags)
 {
         struct ip_vs_conn *cp;
-        static int n_fails = 0;
 
-        cp = kmalloc(sizeof(struct ip_vs_conn), GFP_KERNEL);
+        cp = kmalloc(sizeof(struct ip_vs_conn), GFP_ATOMIC);
         if (cp == NULL) {
-                if (++n_fails < 5)
+                if (net_ratelimit())
                         IP_VS_ERR("ip_vs_conn_new: no memory available.\n");
                 return NULL;
         }
@@ -794,20 +812,24 @@ ip_vs_conn_new(int proto, __u32 caddr, __u16 cport, __u32 vaddr, __u16 vport,
         cp->app_data	   = NULL;
         cp->control	   = NULL;
 	
-	atomic_set(&cp->n_control,0);
-	atomic_set(&cp->refcnt,0);
-	atomic_set(&cp->in_pkts,0);
+	atomic_set(&cp->n_control, 0);
+	atomic_set(&cp->refcnt, 0);
+	atomic_set(&cp->in_pkts, 0);
 
         /*
          *  Hash it in the ip_vs_table
          */
         ip_vs_conn_hash(cp);
 
+        if (flags & IP_VS_CONN_F_NO_CPORT)
+                atomic_inc(&ip_vs_conn_no_cport_cnt);
+
         ip_vs_bind_app(cp);
-        atomic_inc(&cp->refcnt);
         vs_set_state_timeout(cp, IP_VS_S_NONE);
 
-        n_fails = 0;
+        /*  Just increase its refcnt so that it can be put back */
+        atomic_inc(&cp->refcnt);
+
         return cp;
 }
 
@@ -841,7 +863,7 @@ ip_vs_conn_getinfo(char *buffer, char **start, off_t offset, int length)
                         cp = list_entry(e, struct ip_vs_conn, c_list);
 
                         size = sprintf(buffer+len,
-                                       "%s %08X %04X %08X %04X "
+                                       "%-3s %08X %04X %08X %04X "
                                        "%08X %04X %-11s %7lu\n",
                                        vs_proto_name(cp->protocol),
                                        ntohl(cp->caddr), ntohs(cp->cport),
@@ -969,6 +991,12 @@ int ip_vs_conn_init(void)
         if (!ip_vs_table) {
                 return -ENOMEM;
         }
+        IP_VS_INFO("Connection hash table configured "
+                   "(size=%d, memory=%ldKbytes)\n",
+                   IP_VS_CONN_TAB_SIZE,
+                   (long)(IP_VS_CONN_TAB_SIZE*sizeof(struct list_head))/1024);
+        IP_VS_DBG(0, "Each connection entry needs %d bytes at least\n",
+                  sizeof(struct ip_vs_conn));
 
         for(idx = 0; idx < IP_VS_CONN_TAB_SIZE; idx++)  {
                 INIT_LIST_HEAD(&ip_vs_table[idx]);
