@@ -5,10 +5,11 @@
  *              high-performance and highly available server based on a
  *              cluster of servers.
  *
- * Version:     $Id: ip_vs_core.c,v 1.1.1.1 2000/05/19 07:05:08 wensong Exp $
+ * Version:     $Id: ip_vs_core.c,v 1.5 2000/07/06 14:57:09 wensong Exp $
  *
  * Authors:     Wensong Zhang <wensong@iinchina.net>
  *              Peter Kese <peter.kese@ijs.si>
+ *              Julian Anastasov <uli@linux.tu-varna.acad.bg>
  *
  *              This program is free software; you can redistribute it and/or
  *              modify it under the terms of the GNU General Public License
@@ -53,12 +54,16 @@ EXPORT_SYMBOL(register_ip_vs_scheduler);
 EXPORT_SYMBOL(unregister_ip_vs_scheduler);
 EXPORT_SYMBOL(ip_vs_bind_dest);
 EXPORT_SYMBOL(ip_vs_unbind_dest);
+EXPORT_SYMBOL(ip_vs_skb_replace);
+EXPORT_SYMBOL(vs_proto_name);
+EXPORT_SYMBOL(ip_vs_conn_new);
+EXPORT_SYMBOL(ip_vs_conn_in_get);
+EXPORT_SYMBOL(ip_vs_conn_out_get);
+EXPORT_SYMBOL(ip_vs_conn_listen);
+EXPORT_SYMBOL(ip_vs_conn_put);
+#ifdef CONFIG_IP_VS_DEBUG
 EXPORT_SYMBOL(ip_vs_get_debug_level);
-
-/* FIXME:
- *       Many functions here doesn't look good to me.
- *       Rewrite them!!!!
- */
+#endif
 
 
 /* ID used in ICMP lookups */
@@ -96,7 +101,7 @@ static __inline__ int proto_doff(unsigned proto, char *th, unsigned size)
                 }
                 break;
         default:
-                /* 	Other proto: nothing to say, by now :) */
+                /* Other proto: nothing to say, by now :) */
                 ret = 0;
 	}
 	if (ret < 0)
@@ -111,6 +116,8 @@ const char *vs_proto_name(unsigned proto)
 	static char buf[20];
 
         switch (proto) {
+        case IPPROTO_IP:
+                return "IP";
         case IPPROTO_UDP:
                 return "UDP";
         case IPPROTO_TCP:
@@ -123,12 +130,13 @@ const char *vs_proto_name(unsigned proto)
         }
 }
 
+
 /*
  *	Own skb_cow() beast, tweaked for rewriting commonly
- *	used pointers in masq code
+ *	used pointers in the code
  */
-static struct sk_buff *masq_skb_cow(struct sk_buff **skb_p,
-                                    struct iphdr **iph_p, unsigned char **t_p)
+static struct sk_buff *vs_skb_cow(struct sk_buff **skb_p,
+                                  struct iphdr **iph_p, unsigned char **t_p)
 {
 	struct sk_buff *skb=(*skb_p);
 	if (skb_cloned(skb)) {
@@ -169,7 +177,8 @@ ip_vs_sched_persist(struct ip_vs_service *svc, struct iphdr *iph)
 	/* Mask saddr with the netmask to adjust template granularity */
 	snet = iph->saddr & svc->netmask;
 
-	IP_VS_DBG(1, "P-schedule: src %u.%u.%u.%u:%u dest %u.%u.%u.%u:%u mnet %u.%u.%u.%u\n",
+	IP_VS_DBG(1, "P-schedule: src %u.%u.%u.%u:%u dest %u.%u.%u.%u:%u "
+                  "mnet %u.%u.%u.%u\n",
 		  NIPQUAD(iph->saddr), ntohs(portp[0]),
 		  NIPQUAD(iph->daddr), ntohs(portp[1]),
 		  NIPQUAD(snet));
@@ -207,11 +216,12 @@ ip_vs_sched_persist(struct ip_vs_service *svc, struct iphdr *iph)
 				return NULL;
 			}
 
-			/*
-			 * Create a template <caddr,0,vaddr,vport,daddr,dport>
-			 * for non-ftp service, and <caddr,0,vaddr,0,daddr,0>
-			 * for ftp service.
-			 */
+                        /*
+                         * Create a template like <protocol,daddr,0,
+                         * maddr,mport,saddr,sport> for non-ftp service,
+                         * and <protocol,daddr,0,maddr,0,saddr,0>
+                         * for ftp service.
+                         */
 			if (svc->port != FTPPORT)
 				ct = ip_vs_conn_new(iph->protocol,
 						    snet, 0,
@@ -226,7 +236,6 @@ ip_vs_sched_persist(struct ip_vs_service *svc, struct iphdr *iph)
 						    0);
 			if (ct == NULL) {
 				IP_VS_ERR("ip_vs_conn_new template failed\n");
-				read_unlock(&__ip_vs_svc_lock);
 				return NULL;
 			}
 
@@ -235,7 +244,6 @@ ip_vs_sched_persist(struct ip_vs_service *svc, struct iphdr *iph)
 			 */
 			ip_vs_bind_dest(ct, dest);
 			ct->timeout = svc->timeout;
-
 		} else {
 			/*
 			 * Template found and its destination is available.
@@ -249,8 +257,18 @@ ip_vs_sched_persist(struct ip_vs_service *svc, struct iphdr *iph)
 		}
 		dport = dest->port;
 	} else {
-		ct = ip_vs_conn_in_get(iph->protocol, snet, 0,
-				       iph->daddr, 0);
+                /*
+                 * Note: persistent fwmark-based services and persistent
+                 * port zero service are handled here.
+                 * fwmark template: <IPPROTO_IP,daddr,0,maddr,0,saddr,0>
+                 * port zero template: <protocol,daddr,0,maddr,0,saddr,0>
+                 */
+                if (svc->fwmark)
+                        ct = ip_vs_conn_in_get(IPPROTO_IP, snet, 0,
+                                               iph->daddr, 0);
+                else
+                        ct = ip_vs_conn_in_get(iph->protocol, snet, 0,
+                                               iph->daddr, 0);
 
 		if (!ct || !ip_vs_check_template(ct)) {
 			/*
@@ -266,17 +284,23 @@ ip_vs_sched_persist(struct ip_vs_service *svc, struct iphdr *iph)
 				return NULL;
 			}
 
-			/*
-			 * Create a template <caddr,0,vaddr,0,daddr,0>
-			 */
-			ct = ip_vs_conn_new(iph->protocol,
-					    snet, 0,
-					    iph->daddr, 0,
-					    dest->addr, 0,
-					    0);
+                        /*
+                         * Create a template according to the service
+                         */
+                        if (svc->fwmark)
+                                ct = ip_vs_conn_new(IPPROTO_IP,
+                                                    snet, 0,
+                                                    iph->daddr, 0,
+                                                    dest->addr, 0,
+                                                    0);
+                        else
+                                ct = ip_vs_conn_new(iph->protocol,
+                                                    snet, 0,
+                                                    iph->daddr, 0,
+                                                    dest->addr, 0,
+                                                    0);
 			if (ct == NULL) {
 				IP_VS_ERR("ip_vs_conn_new template failed\n");
-				read_unlock(&__ip_vs_svc_lock);
 				return NULL;
 			}
 
@@ -355,7 +379,9 @@ struct ip_vs_conn *ip_vs_schedule(struct ip_vs_service *svc, struct iphdr *iph)
 	portp = (__u16 *)&(((char *)iph)[iph->ihl*4]);
 	if (!svc->fwmark && portp[1] != svc->port) {
 		if (!svc->port)
-			IP_VS_ERR("Schedule: port zero only supported in persistent services, check your ipvs configuration\n");
+			IP_VS_ERR("Schedule: port zero only supported "
+                                  "in persistent services, "
+                                  "check your ipvs configuration\n");
 		return NULL;
 	}
 
@@ -419,7 +445,7 @@ int ip_vs_leave(struct ip_vs_service *svc, struct sk_buff *skb)
 	 */
 	if ((svc->port == FTPPORT) && (portp[1] != FTPPORT)) {
 		atomic_dec(&svc->refcnt);
-		return 0;
+		return NF_ACCEPT;
 	}
 	atomic_dec(&svc->refcnt);
 
@@ -432,7 +458,160 @@ int ip_vs_leave(struct ip_vs_service *svc, struct sk_buff *skb)
 	 */
 	icmp_send(skb, ICMP_DEST_UNREACH, ICMP_PORT_UNREACH, 0);
 	kfree_skb(skb);
-	return -2;
+	return NF_STOLEN;
+}
+
+
+/*
+ *      NAT transmitter (only for outside-to-inside nat forwarding)
+ */
+int ip_vs_nat_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
+{
+        struct rtable *rt;			/* Route to the other host */
+	struct iphdr  *iph = skb->nh.iph;
+	u8     tos = iph->tos;
+	union ip_vs_tphdr h;
+	unsigned short size;
+	int doff;
+	int csum = 0;
+	int csum_ok = 0;
+        int mtu;
+
+	EnterFunction(1);
+
+        h.raw = (char*) iph + iph->ihl * 4;
+	size = ntohs(iph->tot_len) - (iph->ihl * 4);
+	doff = proto_doff(iph->protocol, h.raw, size);
+
+        /*
+         *  Check if it is no_cport connection (VS/NAT only)...
+         */
+        if (cp->flags & IP_VS_CONN_F_NO_CPORT ) {
+                atomic_dec(&ip_vs_conn_no_cport_cnt);
+                ip_vs_conn_unhash(cp);
+                cp->flags &= ~IP_VS_CONN_F_NO_CPORT;
+                cp->cport = h.portp[0];
+                /* hash on new dport */
+                ip_vs_conn_hash(cp);
+
+                IP_VS_DBG(1, "filled cport=%d\n", ntohs(cp->dport));
+        }
+                
+        /*
+         *  Must do TCP/UDP checksum checking before mangling the packet
+         */
+        if ((iph->protocol != IPPROTO_UDP) || (h.uh->check != 0)) {
+                switch (skb->ip_summed) {
+                case CHECKSUM_NONE:
+                        csum = csum_partial(h.raw + doff, size - doff, 0);
+                        csum_ok++;
+                        skb->csum = csum_partial(h.raw , doff, csum);
+
+                case CHECKSUM_HW:
+                        if (csum_tcpudp_magic(iph->saddr, iph->daddr,
+                                              size, iph->protocol, skb->csum))
+                        {
+                                /*  atomic_dec(&svc->refcnt); */
+                                IP_VS_DBG(0, "Incoming failed %s checksum "
+                                          "from %d.%d.%d.%d (size=%d)!\n",
+                                          vs_proto_name(iph->protocol),
+                                          NIPQUAD(iph->saddr),
+                                          size);
+                                kfree_skb(skb);
+                                return -EINVAL;
+                        }
+                default:/* CHECKSUM_UNNECESSARY */
+                }
+        }
+        
+        if (ip_route_output(&rt, cp->daddr, 0, RT_TOS(tos), 0)) {
+		IP_VS_DBG(0, "ip_vs_nat_xmit(): route error, dest: %08X\n",
+                          cp->daddr);
+		goto tx_error_icmp;
+	}
+
+        /* MTU checking */
+        mtu = rt->u.dst.pmtu;
+        if ((skb->len > mtu) && (iph->frag_off&__constant_htons(IP_DF))) {
+                icmp_send(skb, ICMP_DEST_UNREACH,ICMP_FRAG_NEEDED, htonl(mtu));
+                ip_rt_put(rt);
+                IP_VS_DBG(0, "ip_vs_nat_xmit(): frag needed\n");
+                goto tx_error;
+        }
+        
+	/* drop old route */
+        dst_release(skb->dst);
+	skb->dst = &rt->u.dst;
+
+        /* mangle the packet */
+        iph->daddr = cp->daddr;
+        h.portp[1] = cp->dport;
+
+        /*
+         *	Invalidate csum saving if it has ip_vs_app helper
+         */
+        if (cp->app)
+                csum_ok = 0;
+
+        /*
+         *	Attempt ip_vs_app call.
+         *	will fix ip_vs_conn and iph ack_seq stuff
+         */
+        if (ip_vs_app_pkt_in(cp, &skb) != 0) {
+                /*
+                 *	skb has changed, update pointers.
+                 */
+                /*  skb = *skb_p; */
+                iph = skb->nh.iph;
+                h.raw = (char*) iph + iph->ihl*4;
+                size = ntohs(iph->tot_len) - (iph->ihl * 4);
+        }
+
+        /*
+         * Yug! adjust UDP/TCP checksums
+         */
+        /*
+         *	Transport's payload partial csum
+         */
+        if (!csum_ok) {
+                csum = csum_partial(h.raw + doff, size - doff, 0);
+        }
+        skb->csum = csum;
+
+        /*
+         *	Protocol csum
+         */
+        switch (iph->protocol) {
+        case IPPROTO_TCP:
+                h.th->check = 0;
+                h.th->check=csum_tcpudp_magic(iph->saddr, iph->daddr,
+                                              size, iph->protocol,
+                                              csum_partial(h.raw, doff, csum));
+                break;
+        case IPPROTO_UDP:
+                h.uh->check = 0;
+                h.uh->check=csum_tcpudp_magic(iph->saddr, iph->daddr,
+                                              size, iph->protocol,
+                                              csum_partial(h.raw, doff, csum));
+                if (h.uh->check == 0)
+                        h.uh->check = 0xFFFF;
+                break;
+        }
+        ip_send_check(iph);
+
+        IP_VS_DBG(1, "NAT to %u.%u.%u.%u:%d\n",
+                  NIPQUAD(iph->daddr), ntohs(h.portp[1]));
+
+        ip_send(skb);
+
+        LeaveFunction(1);
+	return 0;
+
+  tx_error_icmp:
+	dst_link_failure(skb);
+  tx_error:
+	kfree_skb(skb);
+	return -EINVAL;
 }
 
 
@@ -449,10 +628,9 @@ int ip_vs_leave(struct ip_vs_service *svc, struct sk_buff *skb)
  *     the client without passing the load balancer. This can greatly
  *     increase the scalability of virtual server.
  *   Returns:
- *     if succeeded, return 1; otherwise, return 0.
+ *     if succeeded, return 0; otherwise, return -EINVAL.
  */
-
-int ip_vs_tunnel_xmit(struct sk_buff *skb, __u32 daddr)
+int ip_vs_tunnel_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
 {
 	struct rtable *rt;			/* Route to the other host */
 	struct net_device *tdev;		/* Device to other host */
@@ -461,20 +639,22 @@ int ip_vs_tunnel_xmit(struct sk_buff *skb, __u32 daddr)
 	u16    df = old_iph->frag_off;
 	struct iphdr  *iph;			/* Our new IP header */
 	int    max_headroom;			/* The extra header space needed */
-	u32    dst = daddr;
+	u32    dst = cp->daddr;
 	u32    src = 0;
 	int    mtu;
 
-	EnterFunction;
+	EnterFunction(1);
 
         if (skb->protocol != __constant_htons(ETH_P_IP)) {
-		IP_VS_DBG(0, "ip_vs_tunnel_xmit(): protocol error, ETH_P_IP: %d, skb protocol: %d\n",
-			  __constant_htons(ETH_P_IP),skb->protocol);
+		IP_VS_DBG(0, "ip_vs_tunnel_xmit(): protocol error, "
+                          "ETH_P_IP: %d, skb protocol: %d\n",
+			  __constant_htons(ETH_P_IP), skb->protocol);
 		goto tx_error;
 	}
 
 	if (ip_route_output(&rt, dst, src, RT_TOS(tos), 0)) {
-		IP_VS_DBG(0, "ip_vs_tunnel_xmit(): route error, dst: %08X\n", dst);
+		IP_VS_DBG(0, "ip_vs_tunnel_xmit(): route error, dst: %08X\n",
+                          dst);
 		goto tx_error_icmp;
 	}
 	tdev = rt->u.dst.dev;
@@ -490,8 +670,9 @@ int ip_vs_tunnel_xmit(struct sk_buff *skb, __u32 daddr)
 
 	df |= (old_iph->frag_off&__constant_htons(IP_DF));
 
-	if ((old_iph->frag_off&__constant_htons(IP_DF)) && mtu < ntohs(old_iph->tot_len)) {
-		icmp_send(skb, ICMP_DEST_UNREACH, ICMP_FRAG_NEEDED, htonl(mtu));
+	if ((old_iph->frag_off&__constant_htons(IP_DF))
+            && mtu < ntohs(old_iph->tot_len)) {
+		icmp_send(skb, ICMP_DEST_UNREACH,ICMP_FRAG_NEEDED, htonl(mtu));
 		ip_rt_put(rt);
 		IP_VS_DBG(1, "ip_vs_tunnel_xmit(): frag needed\n");
 		goto tx_error;
@@ -504,12 +685,14 @@ int ip_vs_tunnel_xmit(struct sk_buff *skb, __u32 daddr)
 	 */
 	max_headroom = (((tdev->hard_header_len+15)&~15)+sizeof(struct iphdr));
 
-	if (skb_headroom(skb) < max_headroom || skb_cloned(skb) || skb_shared(skb)) {
-		struct sk_buff *new_skb = skb_realloc_headroom(skb, max_headroom);
+	if (skb_headroom(skb) < max_headroom
+            || skb_cloned(skb) || skb_shared(skb)) {
+		struct sk_buff *new_skb =
+                        skb_realloc_headroom(skb, max_headroom);
 		if (!new_skb) {
 			ip_rt_put(rt);
 			kfree_skb(skb);
-			IP_VS_ERR("ip_vs_tunnel_xmit(): no memory for new_skb\n");
+			IP_VS_ERR("ip_vs_tunnel_xmit(): no memory\n");
 			return -EINVAL;
 		}
 		kfree_skb(skb);
@@ -518,13 +701,18 @@ int ip_vs_tunnel_xmit(struct sk_buff *skb, __u32 daddr)
 
 	skb->nh.raw = skb_push(skb, sizeof(struct iphdr));
 	memset(&(IPCB(skb)->opt), 0, sizeof(IPCB(skb)->opt));
+
+	/* drop old route */
 	dst_release(skb->dst);
 	skb->dst = &rt->u.dst;
 
+        /* drop conntrack */
+/*          nf_conntrack_put(skb->nfct); */
+/*          skb->nfct = NULL; */
+
 	/*
 	 *	Push down and install the IPIP header.
 	 */
-
 	iph			=	skb->nh.iph;
 	iph->version		=	4;
 	iph->ihl		=	sizeof(struct iphdr)>>2;
@@ -538,12 +726,9 @@ int ip_vs_tunnel_xmit(struct sk_buff *skb, __u32 daddr)
 	ip_select_ident(iph, &rt->u.dst);
 	ip_send_check(iph);
 
-        nf_conntrack_put(skb->nfct);
-        skb->nfct = NULL;
-
 	ip_send(skb);
 
-        LeaveFunction;
+        LeaveFunction(1);
         
 	return 0;
 
@@ -558,40 +743,46 @@ int ip_vs_tunnel_xmit(struct sk_buff *skb, __u32 daddr)
 /*
  *      Direct Routing
  */
-int ip_vs_dr_xmit(struct sk_buff *skb, __u32 daddr)
+int ip_vs_dr_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
 {
 	struct rtable *rt;			/* Route to the other host */
 	struct iphdr  *iph = skb->nh.iph;
 	u8     tos = iph->tos;
+        int    mtu;
 
-	EnterFunction;
+	EnterFunction(1);
 
-        if (ip_route_output(&rt, daddr, 0, RT_TOS(tos), 0)) {
-		IP_VS_DBG(0, "ip_vs_dr_xmit(): route error, dest: %08X\n", daddr);
+        if (ip_route_output(&rt, cp->daddr, 0, RT_TOS(tos), 0)) {
+		IP_VS_DBG(0, "ip_vs_dr_xmit(): route error, dest: %08X\n",
+                          cp->daddr);
 		goto tx_error_icmp;
 	}
 
+        /* MTU checking */
+        mtu = rt->u.dst.pmtu;
+        if ((skb->len > mtu) && (iph->frag_off&__constant_htons(IP_DF))) {
+                icmp_send(skb, ICMP_DEST_UNREACH,ICMP_FRAG_NEEDED, htonl(mtu));
+                ip_rt_put(rt);
+                IP_VS_DBG(0, "ip_vs_dr_xmit(): frag needed\n");
+                goto tx_error;
+        }
+
 	/* drop old route */
         dst_release(skb->dst);
 	skb->dst = &rt->u.dst;
 
-	/* we need set the flags to avoid packet recursion, or we can
-	   find other mechanism */
-/*	    IPCB(skb)->flags |= IPSKB_REDIRECTED; */
-/*	    IPCB(skb)->flags |= IPSKB_MASQUERADED; */
-
-        skb->nfcache |= NFC_ALTERED;
-        
-        nf_conntrack_put(skb->nfct);
-        skb->nfct = NULL;
+	/* drop conntrack */
+/*          nf_conntrack_put(skb->nfct); */
+/*          skb->nfct = NULL; */
 
         ip_send(skb);
 
-        LeaveFunction;
+        LeaveFunction(1);
 	return 0;
 
   tx_error_icmp:
 	dst_link_failure(skb);
+  tx_error:
 	kfree_skb(skb);
 	return -EINVAL;
 }
@@ -600,19 +791,23 @@ int ip_vs_dr_xmit(struct sk_buff *skb, __u32 daddr)
 /*
  *      ip_vs_forward forwards the packet through tunneling, direct
  *      routing or local node (passing to the upper layer).
+ *      in the outside-to-inside direction.
  */
-/*
- *  FIXME:  need rewriting ip_vs_forward!!!!!!!!!!!!!!!
- *          rewrite it to put in NetFilter framework naturally.
- */
-
-static int ip_vs_forward(struct sk_buff *skb, struct ip_vs_conn *cp)
+static inline int ip_vs_forward(struct sk_buff *skb, struct ip_vs_conn *cp)
 {
         atomic_inc(&cp->in_pkts);
         
-        switch (IP_VS_VS_FWD(cp)) {
+        switch (IP_VS_FWD_METHOD(cp)) {
+        case IP_VS_CONN_F_MASQ:
+                if (ip_vs_nat_xmit(skb, cp)) {
+                        IP_VS_DBG(2, "nat failed.\n");
+                } else {
+                        IP_VS_DBG(2, "nat succeeded.\n");
+                }
+                return NF_STOLEN;
+                
         case IP_VS_CONN_F_TUNNEL:
-                if (ip_vs_tunnel_xmit(skb, cp->daddr)) {
+                if (ip_vs_tunnel_xmit(skb, cp)) {
                         IP_VS_DBG(2, "tunneling failed.\n");
                 } else {
                         IP_VS_DBG(2, "tunneling succeeded.\n");
@@ -620,7 +815,7 @@ static int ip_vs_forward(struct sk_buff *skb, struct ip_vs_conn *cp)
                 return NF_STOLEN;
                                 
         case IP_VS_CONN_F_DROUTE:
-                if (ip_vs_dr_xmit(skb, cp->daddr)) {
+                if (ip_vs_dr_xmit(skb, cp)) {
                         IP_VS_DBG(2, "direct routing failed.\n");
                 } else {
                         IP_VS_DBG(2, "direct routing succeeded.\n");
@@ -637,7 +832,11 @@ static int ip_vs_forward(struct sk_buff *skb, struct ip_vs_conn *cp)
 
 
 /*
- *	Handle outgoing ICMP messages.
+ *	Handle ICMP messages in the inside-to-outside direction (outgoing).
+ *	Find any that might be relevant, check against existing connections,
+ *	forward to the right destination host if relevant.
+ *	Currently handles error types - unreachable, quench, ttl exceeded.
+ *      (Only used in VS/NAT)
  */
 int ip_vs_out_icmp(struct sk_buff **skb_p)
 {
@@ -689,10 +888,10 @@ int ip_vs_out_icmp(struct sk_buff **skb_p)
 		return NF_DROP;
 	}
 
-
-	IP_VS_DBG(1, "Handling forward ICMP for %08X:%04X -> %08X:%04X\n",
-		  ntohl(ciph->saddr), ntohs(pptr[0]),
-		  ntohl(ciph->daddr), ntohs(pptr[1]));
+	IP_VS_DBG(1, "Handling outgoing ICMP for "
+                  "%u.%u.%u.%u:%d -> %u.%u.%u.%u:%d\n",
+		  NIPQUAD(ciph->saddr), ntohs(pptr[0]),
+		  NIPQUAD(ciph->daddr), ntohs(pptr[1]));
 
 	/* ciph content is actually <protocol, caddr, cport, daddr, dport> */
 	cp = ip_vs_conn_out_get(ciph->protocol, ciph->daddr, pptr[1],
@@ -700,9 +899,9 @@ int ip_vs_out_icmp(struct sk_buff **skb_p)
 	if (cp == NULL)
 		return NF_ACCEPT;
 
-	if (IP_VS_VS_FWD(cp) != 0) {
-		IP_VS_INFO("shouldn't reach here, because the box is on the"
-			   "half connection in the tun/dr module.\n");
+	if (IP_VS_FWD_METHOD(cp) != 0) {
+		IP_VS_ERR("shouldn't reach here, because the box is on the"
+                          "half connection in the tun/dr module.\n");
 	}
 
 	/* Now we do real damage to this packet...! */
@@ -722,16 +921,20 @@ int ip_vs_out_icmp(struct sk_buff **skb_p)
 	icmph->checksum = 0;
 	icmph->checksum = ip_compute_csum((unsigned char *) icmph, len);
 
-
-	IP_VS_DBG(1, "Rewrote forward ICMP to %08X:%04X -> %08X:%04X\n",
-		  ntohl(ciph->saddr), ntohs(pptr[0]),
-		  ntohl(ciph->daddr), ntohs(pptr[1]));
-
+	IP_VS_DBG(1, "Forwarding correct outgoing ICMP to "
+                  "%u.%u.%u.%u:%d -> %u.%u.%u.%u:%d\n",
+		  NIPQUAD(ciph->saddr), ntohs(pptr[0]),
+		  NIPQUAD(ciph->daddr), ntohs(pptr[1]));
 
 	return NF_ACCEPT;
 }
 
 
+/*
+ *	It is hooked at the NF_IP_FORWARD chain, used only for VS/NAT.
+ *	Check if outgoing packet belongs to the established ip_vs_conn,
+ *      rewrite addresses of the packet and send it on its way...
+ */
 static unsigned int ip_vs_out(unsigned int hooknum,
                               struct sk_buff **skb_p,
                               const struct net_device *in,
@@ -742,7 +945,7 @@ static unsigned int ip_vs_out(unsigned int hooknum,
 	struct iphdr	*iph = skb->nh.iph;
 	union ip_vs_tphdr h;
 	struct ip_vs_conn *cp;
-	int		size;
+	int size;
 
 	/*
 	 * 	doff holds transport protocol data offset
@@ -753,18 +956,21 @@ static unsigned int ip_vs_out(unsigned int hooknum,
 	int csum = 0;
 	int csum_ok = 0;
 
-	/*
-	 * We can only masquerade protocols with ports... and hack some ICMPs
-	 */
+	EnterFunction(1);
+
 	h.raw = (char*) iph + iph->ihl * 4;
 	size = ntohs(iph->tot_len) - (iph->ihl * 4);
-
 	doff = proto_doff(iph->protocol, h.raw, size);
-	if (doff <= 0) {
+
+	switch (doff) {
+	case 0:
 		/*
-		 *	Output path: do not pass other IP protos nor
-		 *	invalid packets.
+		 *	Input path: other IP protos Ok, will
+		 *	reach local sockets path.
 		 */
+		return NF_ACCEPT;
+	case -1:
+		IP_VS_DBG(0, "I-pkt invalid packet data size\n");
 		return NF_DROP;
 	}
 
@@ -772,16 +978,33 @@ static unsigned int ip_vs_out(unsigned int hooknum,
 	case IPPROTO_ICMP:
 		return ip_vs_out_icmp(skb_p);
 	case IPPROTO_UDP:
-		if (h.uh->check == 0)
-			/* No UDP checksum */
-			break;
 	case IPPROTO_TCP:
-		IP_VS_DBG(3, "O-pkt: %s size=%d\n",
-                          vs_proto_name(iph->protocol), size);
+		break;
+	default:
+		return NF_ACCEPT;
+	}
 
-		/* Check that the checksum is OK */
-		switch (skb->ip_summed)
-		{
+        IP_VS_DBG(3, "O-pkt: %s size=%d\n",
+                  vs_proto_name(iph->protocol), size);
+                
+        /*
+         *	Now hunt the list to see if we have an old entry
+         */
+        cp = ip_vs_conn_out_get(iph->protocol, iph->saddr, h.portp[0],
+                                iph->daddr, h.portp[1]);
+        if (!cp) {
+                IP_VS_DBG(2, "packet for %s %d.%d.%d.%d:%d "
+                          "continue traversal as normal.\n",
+                          vs_proto_name(iph->protocol),
+                          NIPQUAD(iph->daddr),
+                          ntohs(h.portp[1]));
+                return NF_ACCEPT;
+        }
+
+        /* Check if the checksum is OK, when it is not UDP or check!=0 */
+        if ((iph->protocol != IPPROTO_UDP) || (h.uh->check != 0)) {
+                switch (skb->ip_summed)
+                {
                 case CHECKSUM_NONE:
                         csum = csum_partial(h.raw + doff, size - doff, 0);
                         IP_VS_DBG(3, "O-pkt: %s I-datacsum=%d\n",
@@ -800,34 +1023,15 @@ static unsigned int ip_vs_out(unsigned int hooknum,
                                 return NF_DROP;
                         }
                 default:
-				/* CHECKSUM_UNNECESSARY */
-		}
-		break;
-	default:
-		return NF_DROP;
-	}
+                        /* CHECKSUM_UNNECESSARY */
+                }
+        }
 
-	/*
-	 *	Now hunt the list to see if we have an old entry
-	 */
-	/* h.raw = (char*) iph + iph->ihl * 4; */
-
- 	IP_VS_DBG(2, "Outgoing %s %u.%u.%u.%u:%d->%u.%u.%u.%u:%d\n",
+        IP_VS_DBG(2, "Outgoing %s %u.%u.%u.%u:%d->%u.%u.%u.%u:%d\n",
                   vs_proto_name(iph->protocol),
                   NIPQUAD(iph->saddr), ntohs(h.portp[0]),
                   NIPQUAD(iph->daddr), ntohs(h.portp[1]));
 
-        cp = ip_vs_conn_out_get(iph->protocol, iph->saddr, h.portp[0],
-                                iph->daddr, h.portp[1]);
-        if (!cp) {
-                IP_VS_DBG(2, "packet for %s %d.%d.%d.%d:%d "
-                          "continue traversal as normal.\n",
-                          vs_proto_name(iph->protocol),
-                          NIPQUAD(iph->daddr),
-                          ntohs(h.portp[1]));
-                return NF_ACCEPT;
-        }
-
  	/*
  	 *	Change the fragments origin
  	 */
@@ -863,7 +1067,6 @@ static unsigned int ip_vs_out(unsigned int hooknum,
  	/*
  	 *	Adjust packet accordingly to protocol
  	 */
-
 	/*
 	 *	Transport's payload partial csum
 	 */
@@ -904,12 +1107,14 @@ static unsigned int ip_vs_out(unsigned int hooknum,
 
 	vs_set_state(cp, VS_STATE_OUTPUT, iph, h.portp);
 	ip_vs_conn_put(cp);
+        
+        LeaveFunction(1);
 	return NF_ACCEPT;
 }
 
 
 /*
- *	Handle ICMP messages in reverse direction (incoming).
+ *	Handle ICMP messages in the outside-to-inside direction (incoming).
  *	Find any that might be relevant, check against existing connections,
  *	forward to the right destination host if relevant.
  *	Currently handles error types - unreachable, quench, ttl exceeded
@@ -921,15 +1126,16 @@ int ip_vs_in_icmp(struct sk_buff **skb_p)
 	struct icmphdr  *icmph = (struct icmphdr *)((char *)iph+(iph->ihl<<2));
 	struct iphdr    *ciph;	/* The ip header contained within the ICMP */
 	__u16	        *pptr;	/* port numbers from TCP/UDP contained header */
-	struct ip_vs_conn *cp;
 	unsigned short   len   = ntohs(iph->tot_len) - (iph->ihl * 4);
+	struct ip_vs_conn *cp;
+        struct rtable *rt;			/* Route to the other host */
+        u8     tos = iph->tos;
+        int    mtu;
 
-
- 	IP_VS_DBG(2, "icmp in/rev (%d,%d) %u.%u.%u.%u -> %u.%u.%u.%u\n",
+ 	IP_VS_DBG(2, "icmp in (%d,%d) %u.%u.%u.%u -> %u.%u.%u.%u\n",
                   icmph->type, ntohs(icmp_id(icmph)),
                   NIPQUAD(iph->saddr), NIPQUAD(iph->daddr));
 
-
         if ((icmph->type != ICMP_DEST_UNREACH) &&
             (icmph->type != ICMP_SOURCE_QUENCH) &&
             (icmph->type != ICMP_TIME_EXCEEDED))
@@ -950,16 +1156,17 @@ int ip_vs_in_icmp(struct sk_buff **skb_p)
 	if (ip_compute_csum((unsigned char *) icmph, len))
 	{
 		/* Failed checksum! */
-		IP_VS_ERR( "reverse ICMP: failed checksum from %d.%d.%d.%d!\n",
-                           NIPQUAD(iph->saddr));
+		IP_VS_ERR("incoming ICMP: failed checksum from %d.%d.%d.%d!\n",
+                          NIPQUAD(iph->saddr));
 		return NF_DROP;
 	}
 
         pptr = (__u16 *)&(((char *)ciph)[ciph->ihl*4]);
 
- 	IP_VS_DBG(2, "Handling reverse ICMP for %08X:%04X -> %08X:%04X\n",
-                  ntohl(ciph->saddr), ntohs(pptr[0]),
-                  ntohl(ciph->daddr), ntohs(pptr[1]));
+ 	IP_VS_DBG(2, "Handling incoming ICMP for "
+                  "%u.%u.%u.%u:%d -> %u.%u.%u.%u:%d\n",
+                  NIPQUAD(ciph->saddr), ntohs(pptr[0]),
+                  NIPQUAD(ciph->daddr), ntohs(pptr[1]));
 
 	/* This is pretty much what ip_vs_conn_in_get_iph() does,
            except paracp are wrong way round */
@@ -969,13 +1176,17 @@ int ip_vs_in_icmp(struct sk_buff **skb_p)
 	if (cp == NULL)
 		return NF_ACCEPT;
 
-        if (IP_VS_VS_FWD(cp) != 0) {
+        /* The ICMP packet for VS/TUN, VS/DR and LOCALNODE will be
+           forwarded directly here */
+        if (IP_VS_FWD_METHOD(cp) != IP_VS_CONN_F_MASQ) {
                 int ret = ip_vs_forward(skb, cp);
                 ip_vs_conn_put(cp);
                 return ret;
         }
 
-	if ((skb=masq_skb_cow(skb_p, &iph, (unsigned char**)&icmph)) == NULL) {
+	/* The ICMP packet for VS/NAT must be written to correct addresses
+           before being forwarded to the right server */
+        if ((skb=vs_skb_cow(skb_p, &iph, (unsigned char**)&icmph)) == NULL) {
 		__ip_vs_conn_put(cp);
 		return NF_DROP;
 	}
@@ -1000,22 +1211,50 @@ int ip_vs_in_icmp(struct sk_buff **skb_p)
 	icmph->checksum = 0;
 	icmph->checksum = ip_compute_csum((unsigned char *) icmph, len);
 
- 	IP_VS_DBG(2, "Rewrote reverse ICMP to %08X:%04X -> %08X:%04X\n",
-                  ntohl(ciph->saddr), ntohs(pptr[0]),
-                  ntohl(ciph->daddr), ntohs(pptr[1]));
+ 	IP_VS_DBG(2, "Forwarding incoming ICMP to "
+                  "%u.%u.%u.%u:%d -> %u.%u.%u.%u:%d\n",
+                  NIPQUAD(ciph->saddr), ntohs(pptr[0]),
+                  NIPQUAD(ciph->daddr), ntohs(pptr[1]));
 
-	return NF_ACCEPT;
+        /* send the mangled packet immediately here */
+        if (ip_route_output(&rt, cp->daddr, 0, RT_TOS(tos), 0)) {
+                IP_VS_DBG(0, "ip_vs_in_imcp(): route error, dest: %08X\n",
+                          cp->daddr);
+                goto tx_error_icmp;
+        }
+
+        /* MTU checking, is it necessary here??? */
+        mtu = rt->u.dst.pmtu;
+        if ((skb->len > mtu) && (iph->frag_off&__constant_htons(IP_DF))) {
+                icmp_send(skb, ICMP_DEST_UNREACH,ICMP_FRAG_NEEDED, htonl(mtu));
+                ip_rt_put(rt);
+                IP_VS_DBG(0, "ip_vs_in_icmp(): frag needed\n");
+                goto tx_error;
+        }
+
+        /* drop old route */
+        dst_release(skb->dst);
+        skb->dst = &rt->u.dst;
+
+/*          skb->nfcache |= NFC_ALTERED; */
+/*          nf_conntrack_put(skb->nfct); */
+/*          skb->nfct = NULL; */
+
+        ip_send(skb);
+        return NF_STOLEN;
+
+  tx_error_icmp:
+	dst_link_failure(skb);
+  tx_error:
+	dev_kfree_skb(skb);
+	return NF_STOLEN;
 }
 
 
- /*
-  *	Check if it's for virtual services, look it up,
-  *	and send it on its way...
-  *
-  *	Better not have many hosts using the designated portrange
-  *	as 'normal' ports, or you'll be spending many time in
-  *	this function.
-  */
+/*
+ *	Check if it's for virtual services, look it up,
+ *	and send it on its way...
+ */
 static unsigned int ip_vs_in(unsigned int hooknum,
 			     struct sk_buff **skb_p,
 			     const struct net_device *in,
@@ -1028,9 +1267,7 @@ static unsigned int ip_vs_in(unsigned int hooknum,
 	struct ip_vs_conn *cp;
 	unsigned short size;
 	int doff = 0;
-	int csum = 0;
-	int csum_ok = 0;
-	struct ip_vs_service *svc = NULL;
+	struct ip_vs_service *svc;
 
 	/*
 	 *	Big tappo: only PACKET_HOST (nor loopback neither mcasts)
@@ -1069,46 +1306,6 @@ static unsigned int ip_vs_in(unsigned int hooknum,
 		return(ip_vs_in_icmp(skb_p));
 	case IPPROTO_TCP:
 	case IPPROTO_UDP:
-		svc = ip_vs_lookup_service(skb->nfmark, iph->protocol,
-                                           iph->daddr, h.portp[1]);
-		if (!svc) {
-                        IP_VS_DBG(3, "packet for %s %d.%d.%d.%d:%d "
-                                  "continue traversal as normal.\n",
-                                  vs_proto_name(iph->protocol),
-                                  NIPQUAD(iph->daddr),
-                                  ntohs(h.portp[1]));
-			return NF_ACCEPT;
-                }
-                
-		/*
-		 *  FIXME: whether checksum checking is necessary??
-		 */
-
-		/* Check that the checksum is OK */
-		if ((iph->protocol == IPPROTO_UDP) && (h.uh->check == 0))
-			/* No UDP checksum */
-			break;
-
-		switch (skb->ip_summed) {
-		case CHECKSUM_NONE:
-			csum = csum_partial(h.raw + doff, size - doff, 0);
-			csum_ok++;
-			skb->csum = csum_partial(h.raw , doff, csum);
-
-		case CHECKSUM_HW:
-			if (csum_tcpudp_magic(iph->saddr, iph->daddr,
-					      size, iph->protocol, skb->csum))
-			{
-				atomic_dec(&svc->refcnt);
-				IP_VS_DBG(0, "Incoming failed %s checksum "
-                                          "from %d.%d.%d.%d (size=%d)!\n",
-					  vs_proto_name(iph->protocol),
-					  NIPQUAD(iph->saddr),
-					  size);
-				return NF_DROP;
-			}
-		default:/* CHECKSUM_UNNECESSARY */
-		}
 		break;
 	default:
 		return NF_ACCEPT;
@@ -1133,14 +1330,16 @@ static unsigned int ip_vs_in(unsigned int hooknum,
 		 * If the dest is not avaiable, don't restart the timer
 		 * of the packet, but silently drop it.
 		 */
-		atomic_dec(&svc->refcnt);
 		add_sltimer(&cp->timer);
 		__ip_vs_conn_put(cp);
 		return NF_DROP;
 	}
 
-	if (!cp &&
-	    (h.th->syn || (iph->protocol!=IPPROTO_TCP)) && svc) {
+	if (!cp
+            && (h.th->syn || (iph->protocol!=IPPROTO_TCP))
+            && (svc = ip_vs_lookup_service(skb->nfmark, iph->protocol,
+                                           iph->daddr, h.portp[1]))) {
+                
 		if (ip_vs_todrop()) {
 			/*
 			 * It seems that we are very loaded.
@@ -1149,6 +1348,7 @@ static unsigned int ip_vs_in(unsigned int hooknum,
 			atomic_dec(&svc->refcnt);
 			return NF_DROP;
 		}
+                
 		/*
 		 * Let the virtual server select a real server
 		 * for the incomming connection, and create a
@@ -1157,118 +1357,28 @@ static unsigned int ip_vs_in(unsigned int hooknum,
 		cp = ip_vs_schedule(svc, iph);
 		if (!cp)
 			return ip_vs_leave(svc, skb);
+                atomic_dec(&svc->refcnt);
 	}
 
-	if (svc) atomic_dec(&svc->refcnt);
+	if (cp != NULL) {
+                int ret;
 
-	if (cp != NULL)
-	{
-		if (IP_VS_VS_FWD(cp) != 0) {
-			int ret;
-
-			/*
-			 *  Sorry for setting state of conn entry so early
-			 *  no matter whether the packet is forwarded
-			 *  successfully or not, because ip_vs_forward may
-			 *  have already released the skb. Although it
-			 *  brokes the original sematics, it won't lead to
-			 *  serious errors. We look forward to fixing it
-			 *  under the Rusty's netfilter framework both for
-			 *  correctness and modularization.
-			 */
-			vs_set_state(cp, VS_STATE_INPUT, iph, h.portp); 
-
-			ret = ip_vs_forward(skb, cp);
-			ip_vs_conn_put(cp);
-			return ret;
-		}
-
-		IP_VS_DBG(1, "masquerading packet...\n");
-
-		/*
-		 * FIXME: is it necessary to do a copy on write here??
-		 */
-		if ((skb=masq_skb_cow(skb_p, &iph, &h.raw)) == NULL) {
-			ip_vs_conn_put(cp);
-			return NF_DROP;
-		}
-
-		iph->daddr = cp->daddr;
-		h.portp[1] = cp->dport;
-
-		/*
-		 *	Invalidate csum saving if tunnel has connection helper
-		 */
-		if (cp->app)
-			csum_ok = 0;
-
-		/*
-		 *	Attempt ip_vs_app call.
-		 *	will fix ip_vs and iph ack_seq stuff
-		 */
-
-		if (ip_vs_app_pkt_in(cp, skb_p) != 0) {
-			/*
-			 *	skb has changed, update pointers.
-			 */
-			skb = *skb_p;
-			iph = skb->nh.iph;
-			h.raw = (char*) iph + iph->ihl*4;
-			size = ntohs(iph->tot_len) - (iph->ihl * 4);
-		}
-
-		/*
-		 * Yug! adjust UDP/TCP checksums
-		 */
-		/*
-		 *	Transport's payload partial csum
-		 */
-
-		if (!csum_ok) {
-			csum = csum_partial(h.raw + doff, size - doff, 0);
-		}
-		skb->csum = csum;
-
-		/*
-		 *	Protocol csum
-		 */
-		switch (iph->protocol) {
-		case IPPROTO_TCP:
-			h.th->check = 0;
-			h.th->check=csum_tcpudp_magic(iph->saddr, iph->daddr,
-						      size, iph->protocol,
-						      csum_partial(h.raw, doff, csum));
-			break;
-		case IPPROTO_UDP:
-			h.uh->check = 0;
-			h.uh->check=csum_tcpudp_magic(iph->saddr, iph->daddr,
-						      size, iph->protocol,
-						      csum_partial(h.raw, doff, csum));
-			if (h.uh->check == 0)
-				h.uh->check = 0xFFFF;
-			break;
-		}
-		ip_send_check(iph);
-
-		IP_VS_DBG(1, "NAT to %u.%u.%u.%u:%d\n",
-			  NIPQUAD(iph->daddr), ntohs(h.portp[1]));
-
-		vs_set_state(cp, VS_STATE_INPUT, iph, h.portp);
-		ip_vs_conn_put(cp);
-
-                /*
-                  FIXME: the rewritten packet will be caught by our
-                   ip_vs_out hook, there will be unnecessary checking
-                   (such as connection hash table lookup, then not hit).
-                   There are two solutions:
-                   1. ip_send packet directly
-                   2. mark the packet to skip other hooks?
-                   which one?
-                */
-                   
-		/* Let rewritten packet continue traversal as normal */
-		return NF_ACCEPT;
+                vs_set_state(cp, VS_STATE_INPUT, iph, h.portp); 
+                ret = ip_vs_forward(skb, cp);
+                ip_vs_conn_put(cp);
+                return ret;
 	}
+
+#if 0000
+	/*
+          This code is going to die!
+          Warning, svc==0 !
+          TCP packets with SYN=0 from our svc => deliver them
+          locally (don't slowdown the packets not for our services).
+          Let the local host reject them.
+          LVS TCP with SYN=1 are rejected from ip_vs_leave
+        */
+
 	if (svc) {
 		/*
 		 * Drop packet if it is for virtual service but no entry
@@ -1276,12 +1386,19 @@ static unsigned int ip_vs_in(unsigned int hooknum,
 		 * packet to clients if it is not RST or it is not TCP.
 		 */
 		if (!h.th->rst || iph->protocol != IPPROTO_TCP) {
-			icmp_send(skb, ICMP_DEST_UNREACH, ICMP_PORT_UNREACH, 0);
+			icmp_send(skb, ICMP_DEST_UNREACH,ICMP_PORT_UNREACH, 0);
 		}
 		return NF_DROP;
 	}
+#endif
 
 	/* sorry, all this trouble for a no-hit :) */
+        IP_VS_DBG(3, "packet for %s %d.%d.%d.%d:%d continue traversal "
+                  "as normal.\n",
+                  vs_proto_name(iph->protocol),
+                  NIPQUAD(iph->daddr),
+                  ntohs(h.portp[1]));
+
 	return NF_ACCEPT;
 }
 
@@ -1290,7 +1407,7 @@ static unsigned int ip_vs_in(unsigned int hooknum,
    or VS/NAT(change destination) */
 static struct nf_hook_ops ip_vs_in_ops = {
         { NULL, NULL },
-        ip_vs_in, PF_INET, NF_IP_PRE_ROUTING, -100
+        ip_vs_in, PF_INET, /*  NF_IP_PRE_ROUTING */NF_IP_LOCAL_IN, 100
 };
 
 /* After packet filtering, change source only for VS/NAT */
@@ -1299,6 +1416,7 @@ static struct nf_hook_ops ip_vs_out_ops = {
         ip_vs_out, PF_INET, /*  NF_IP_POST_ROUTING */NF_IP_FORWARD, 100
 };
 
+
 /*
  *	Initialize IP Virtual Server
  */
@@ -1310,9 +1428,7 @@ int __init ip_vs_init(void)
 {
 	int ret;
 
-        IP_VS_INFO("Initializing...\n");
-
-	EnterFunction;
+	EnterFunction(1);
 
         ret = ip_vs_control_init();
 	if (ret < 0) {
@@ -1330,7 +1446,7 @@ int __init ip_vs_init(void)
 
 	ret = ip_vs_app_init();
 	if (ret < 0) {
-		IP_VS_ERR("can't setup connection table.\n");
+		IP_VS_ERR("can't setup application helper.\n");
 		goto cleanup_conn;
 	}
 
@@ -1358,9 +1474,8 @@ int __init ip_vs_init(void)
 	ip_vs_wlc_init();
 #endif
 
-        LeaveFunction;
+        LeaveFunction(1);
         
-	/*  MOD_INC_USE_COUNT; */
 	return ret;
 
   cleanup_inops:
@@ -1385,7 +1500,6 @@ int ip_vs_cleanup(void)
 	ip_vs_conn_cleanup();
 	ip_vs_sltimer_cleanup();
 	ip_vs_control_cleanup();
-	/*  MOD_DEC_USE_COUNT; */
 	return 0;
 }
 
