@@ -5,7 +5,7 @@
  *              high-performance and highly available server based on a
  *              cluster of servers.
  *
- * Version:     $Id: ip_vs_ctl.c,v 1.1.1.1 2000/05/20 07:39:10 wensong Exp $
+ * Version:     $Id: ip_vs_ctl.c,v 1.3 2000/07/06 14:57:09 wensong Exp $
  *
  * Authors:     Wensong Zhang <wensong@iinchina.net>
  *              Peter Kese <peter.kese@ijs.si>
@@ -43,23 +43,28 @@
 #include "ip_vs.h"
 //#include <net/ip_vs.h>
 
+/* spinlock for do_ip_vs_set_ctl */
 spinlock_t __ip_vs_mutex = SPIN_LOCK_UNLOCKED;
 
 /* lock for service table */
 rwlock_t __ip_vs_svc_lock = RW_LOCK_UNLOCKED;
+
 /* lock for table with the real services */
 rwlock_t __ip_vs_rs_lock = RW_LOCK_UNLOCKED;
 
+/* lock for state and timeout tables */
+rwlock_t __ip_vs_securetcp_lock = RW_LOCK_UNLOCKED;
+
+/* lock for drop entry handling */
 spinlock_t __ip_vs_dropentry_lock = SPIN_LOCK_UNLOCKED;
 
+/* lock for drop packet handling */
 spinlock_t __ip_vs_droppacket_lock = SPIN_LOCK_UNLOCKED;
+
 int ip_vs_drop_rate = 0;
 int ip_vs_drop_counter = 0;
 atomic_t ip_vs_dropentry = ATOMIC_INIT(0);
 
-/* lock for state and timeout tables */
-rwlock_t __ip_vs_securetcp_lock = RW_LOCK_UNLOCKED;
-
 /* sysctl variables */
 int sysctl_ip_vs_debug_level = 0;
 int sysctl_ip_vs_drop_entry = 0;
@@ -77,7 +82,6 @@ int ip_vs_get_debug_level(void)
 /*
  *  update_defense_level is called from timer bh and from sysctl.
  */
-
 void update_defense_level(void)
 {
 	int ip_vs_amem = nr_free_pages() + atomic_read(&page_cache_size) +
@@ -200,6 +204,13 @@ struct list_head ip_vs_rtable[IP_VS_RTAB_SIZE];
  */
 LIST_HEAD(ip_vs_dest_trash);
 
+/*
+ * FTP & NULL virtual service counters
+ */
+atomic_t ip_vs_ftpsvc_counter = ATOMIC_INIT(0);
+atomic_t ip_vs_nullsvc_counter = ATOMIC_INIT(0);
+
+
 /*
  *  Returns hash value for virtual service
  */
@@ -362,8 +373,10 @@ ip_vs_lookup_service(__u32 fwmark, __u16 protocol, __u32 vaddr, __u16 vport)
 		 */
 		svc = __ip_vs_lookup_service(protocol, vaddr, vport);
 
-		if (svc == NULL && protocol == IPPROTO_TCP &&
-                    (vport == FTPDATA || ntohs(vport) >= PROT_SOCK)) {
+		if (svc == NULL
+                    && protocol == IPPROTO_TCP
+                    && atomic_read(&ip_vs_ftpsvc_counter)
+                    && (vport == FTPDATA || ntohs(vport) >= PROT_SOCK)) {
 			/*
 			 * Check if ftp service entry exists, the packet
 			 * might belong to FTP data connections.
@@ -371,7 +384,8 @@ ip_vs_lookup_service(__u32 fwmark, __u16 protocol, __u32 vaddr, __u16 vport)
 			svc = __ip_vs_lookup_service(protocol, vaddr, FTPPORT);
 		}
 
-		if (svc == NULL) {
+		if (svc == NULL
+                    && atomic_read(&ip_vs_nullsvc_counter)) {
 			/*
 			 * Check if the catch-all port (port zero) exists
 			 */
@@ -381,8 +395,8 @@ ip_vs_lookup_service(__u32 fwmark, __u16 protocol, __u32 vaddr, __u16 vport)
 
 	read_unlock(&__ip_vs_svc_lock);
 
-        IP_VS_DBG(2, "lookup_service fwm %d %d %u.%u.%u.%u:%u %s\n",
-                  fwmark, ntohs(protocol),
+        IP_VS_DBG(2, "lookup_service fwm %d %s %u.%u.%u.%u:%u %s\n",
+                  fwmark, vs_proto_name(protocol),
                   NIPQUAD(vaddr), ntohs(vport),
                   svc?"hit":"not hit");
 
@@ -575,7 +589,6 @@ void __ip_vs_update_dest(struct ip_vs_service *svc,
 	atomic_set(&dest->weight, ur->weight);
 	conn_flags = ur->conn_flags;
 
-	conn_flags |= IP_VS_CONN_F_VS;
 	conn_flags |= IP_VS_CONN_F_INACTIVE;
 
 	/*
@@ -618,7 +631,7 @@ ip_vs_new_dest(struct ip_vs_service *svc, struct ip_vs_rule_user *ur)
 {
 	struct ip_vs_dest *dest;
 
-	EnterFunction;
+	EnterFunction(1);
 
 	dest = (struct ip_vs_dest*) kmalloc(sizeof(struct ip_vs_dest),
 					GFP_ATOMIC);
@@ -642,7 +655,7 @@ ip_vs_new_dest(struct ip_vs_service *svc, struct ip_vs_rule_user *ur)
 	INIT_LIST_HEAD(&dest->d_list);
 	__ip_vs_update_dest(svc, dest, ur);
 
-	LeaveFunction;
+	LeaveFunction(1);
 
 	return dest;
 }
@@ -657,7 +670,7 @@ int ip_vs_add_dest(struct ip_vs_service *svc, struct ip_vs_rule_user *ur)
 	__u32 daddr = ur->daddr;
 	__u16 dport = ur->dport;
 
-	EnterFunction;
+	EnterFunction(1);
 
 	if (ur->weight < 0) {
 		IP_VS_ERR("ip_vs_add_dest(): server weight less than zero\n");
@@ -731,7 +744,7 @@ int ip_vs_add_dest(struct ip_vs_service *svc, struct ip_vs_rule_user *ur)
 
 	write_unlock_bh(&__ip_vs_svc_lock);
 
-        LeaveFunction;
+        LeaveFunction(1);
         
 	return 0;
 }
@@ -746,7 +759,7 @@ int ip_vs_edit_dest(struct ip_vs_service *svc, struct ip_vs_rule_user *ur)
 	__u32 daddr = ur->daddr;
 	__u16 dport = ur->dport;
 
-	EnterFunction;
+	EnterFunction(1);
 
 	if (ur->weight < 0) {
 		IP_VS_ERR("ip_vs_add_dest(): server weight less than zero\n");
@@ -764,7 +777,7 @@ int ip_vs_edit_dest(struct ip_vs_service *svc, struct ip_vs_rule_user *ur)
 
 	__ip_vs_update_dest(svc, dest, ur);
 
-        LeaveFunction;
+        LeaveFunction(1);
         
 	return 0;
 }
@@ -827,7 +840,7 @@ int ip_vs_del_dest(struct ip_vs_service *svc, struct ip_vs_rule_user *ur)
 	__u32 daddr = ur->daddr;
 	__u16 dport = ur->dport;
 
-	EnterFunction;
+	EnterFunction(1);
 
 	/*
 	 *    Lookup the destination list
@@ -856,7 +869,7 @@ int ip_vs_del_dest(struct ip_vs_service *svc, struct ip_vs_rule_user *ur)
 	 */
 	__ip_vs_del_dest(dest);
 
-        LeaveFunction;
+        LeaveFunction(1);
 
 	return 0;
 }
@@ -919,7 +932,7 @@ int ip_vs_add_service(struct ip_vs_rule_user *ur)
 	}
 
 	/*
-	 *	First svc user: the table with the virtual services
+	 *    First svc user: the table with the virtual services
 	 */
 	atomic_set(&svc->refcnt,1);
 
@@ -930,6 +943,14 @@ int ip_vs_add_service(struct ip_vs_rule_user *ur)
 	ip_vs_svc_hash(svc);
 	write_unlock_bh(&__ip_vs_svc_lock);
 
+        /*
+         *    Update the virtual service counters
+         */
+        if (vport == FTPPORT)
+                atomic_inc(&ip_vs_ftpsvc_counter);
+        else if (vport == 0)
+                atomic_inc(&ip_vs_nullsvc_counter);
+
 	return 0;
 
   out_err:
@@ -1015,7 +1036,8 @@ int __ip_vs_del_service(struct ip_vs_service *svc)
 	 */
 	old_sched = svc->scheduler;
 	ip_vs_unbind_scheduler(svc);
-	if (old_sched) __MOD_DEC_USE_COUNT(old_sched->module);
+	if (old_sched)
+                __MOD_DEC_USE_COUNT(old_sched->module);
 
 	/*
 	 *    Unlink the whole destination list
@@ -1023,11 +1045,19 @@ int __ip_vs_del_service(struct ip_vs_service *svc)
 	l = &svc->destinations;
 	while (l->next != l) {
 		dest = list_entry(l->next, struct ip_vs_dest, n_list);
-		__ip_vs_unlink_dest(svc,dest,0);
+		__ip_vs_unlink_dest(svc, dest, 0);
 		__ip_vs_del_dest(dest);
 	}
 
-	/*
+        /*
+         *    Update the virtual service counters
+         */
+        if (svc->port == FTPPORT)
+                atomic_dec(&ip_vs_ftpsvc_counter);
+        else if (svc->port == 0)
+                atomic_dec(&ip_vs_nullsvc_counter);
+
+        /*
 	 * Free the service
 	 */
 	kfree_s(svc, sizeof(struct ip_vs_service));
@@ -1424,7 +1454,8 @@ do_ip_vs_set_ctl(struct sock *sk, int cmd, void *user, unsigned int len)
 	else
 		svc = __ip_vs_lookup_svc_fwm(urule->vfwmark);
 
-	if (svc) atomic_dec(&svc->refcnt);
+	if (svc)
+                atomic_dec(&svc->refcnt);
 
 	switch (cmd) {
 	case IP_VS_SO_SET_ADD:
@@ -1510,7 +1541,7 @@ int ip_vs_control_init(void)
         int ret;
         int idx;
 
-        EnterFunction;
+        EnterFunction(1);
         
  	ret = nf_register_sockopt(&ip_vs_sockopts);
 	if (ret) {
@@ -1534,15 +1565,15 @@ int ip_vs_control_init(void)
                 INIT_LIST_HEAD(&ip_vs_rtable[idx]);
         }
 
-        LeaveFunction;
+        LeaveFunction(1);
         return 0;
 }
 
 void ip_vs_control_cleanup(void)
 {
-        EnterFunction;
+        EnterFunction(1);
         unregister_sysctl_table(ipv4_vs_table.sysctl_header);
         proc_net_remove("ip_vs");
         nf_unregister_sockopt(&ip_vs_sockopts);
-        LeaveFunction;
+        LeaveFunction(1);
 }
