@@ -5,7 +5,7 @@
  *              high-performance and highly available server based on a
  *              cluster of servers.
  *
- * Version:     $Id: ip_vs_sched.c,v 1.1.1.1 2000/05/18 00:24:16 wensong Exp $
+ * Version:     $Id: ip_vs_sched.c,v 1.3 2000/07/06 14:57:09 wensong Exp $
  *
  * Authors:     Wensong Zhang <wensong@iinchina.net>
  *              Peter Kese <peter.kese@ijs.si>
@@ -29,7 +29,7 @@
 #include <linux/errno.h>
 #include <linux/sched.h>
 #include <linux/spinlock.h>
-#include <linux/netdevice.h>
+#include <asm/softirq.h>                /* for local_bh_* */
 #include <asm/string.h>
 #ifdef CONFIG_KMOD
 #include <linux/kmod.h>
@@ -49,6 +49,7 @@ LIST_HEAD(ip_vs_schedulers);
 /* lock for service table */
 rwlock_t __ip_vs_sched_lock = RW_LOCK_UNLOCKED;
 
+
 /*
  *  Register a scheduler in the scheduler list
  */
@@ -201,7 +202,8 @@ struct ip_vs_scheduler *ip_vs_sched_getbyname(const char *sched_name)
 	struct ip_vs_scheduler *sched;
 	struct list_head *l, *e;
 
-	IP_VS_DBG(1, "ip_vs_sched_getbyname(): sched_name \"%s\"\n", sched_name);
+	IP_VS_DBG(1, "ip_vs_sched_getbyname(): sched_name \"%s\"\n",
+                  sched_name);
 	
 	l = &ip_vs_schedulers;
 
@@ -211,12 +213,12 @@ struct ip_vs_scheduler *ip_vs_sched_getbyname(const char *sched_name)
 		sched = list_entry(e, struct ip_vs_scheduler, n_list);
 		
 		/*
-		** Test and MOD_INC_USE_COUNT atomically
-		*/
+                 * Test and MOD_INC_USE_COUNT atomically
+                 */
 		if (!try_inc_mod_count(sched->module)) {
 			/*
-			** This scheduler is just deleted
-			*/
+                         * This scheduler is just deleted
+                         */
 			continue;
 		}
 		if (strcmp(sched_name, sched->name)==0) {
@@ -249,7 +251,7 @@ struct ip_vs_scheduler *ip_vs_lookup_scheduler(const char *sched_name)
 	 */
 	if (sched == NULL) {
 		char module_name[IP_VS_SCHEDNAME_MAXLEN+8];
-		sprintf(module_name,"ip_vs_%s",sched_name);
+		sprintf(module_name,"ip_vs_%s", sched_name);
 #ifdef CONFIG_KMOD
 		request_module(module_name);
 #endif /* CONFIG_KMOD */
