@@ -1,7 +1,7 @@
 /*
  * IPVS:        Weighted Round-Robin Scheduling module
  *
- * Version:     $Id: ip_vs_wrr.c,v 1.1.1.1 2000/05/19 07:49:09 wensong Exp $
+ * Version:     $Id: ip_vs_wrr.c,v 1.3 2000/07/06 14:57:09 wensong Exp $
  *
  * Authors:     Wensong Zhang <wensong@iinchina.net>
  *
@@ -115,11 +115,12 @@ static struct ip_vs_dest* ip_vs_wrr_schedule(struct ip_vs_service *svc)
 		if (mark->cl == &svc->destinations) {
 			/* it is at the head of the destination list */
 			
-			if (mark->cl == mark->cl->next)
+			if (mark->cl == mark->cl->next) {
 				/* no dest entry */
 				write_unlock(&svc->sched_lock);
-				return NULL;
-
+                                return NULL;
+                        }
+                        
 			mark->cl = svc->destinations.next;
 			mark->cw--;
 			if (mark->cw <= 0) {
