@@ -1,7 +1,7 @@
 /*
  *      ipvsadm - IP Virtual Server ADMinistration program
  *
- *      Version: $Id: ipvsadm.c,v 1.11 2000/05/25 13:23:33 wensong Exp $
+ *      Version: $Id: ipvsadm.c,v 1.2 2000/07/06 14:57:09 wensong Exp $
  *
  *      Authors: Wensong Zhang <wensong@iinchina.net>
  *               Peter Kese <peter.kese@ijs.si>
@@ -188,17 +188,16 @@ int main(int argc, char **argv)
 }
 
 
-int process_options(int argc, char **argv, int reading_stdin)
+#ifdef HAVE_POPT
+
+int parse_options(int argc, char **argv, int reading_stdin,
+                  struct ip_vs_rule_user *ur, int *op,
+                  unsigned int *options, unsigned int *format)
 {
-        struct ip_vs_rule_user rule;
-        int sockfd, op=IP_VS_SO_SET_NONE;
-        unsigned int options=OPT_NONE;
-        unsigned int format=FMT_NONE;
         int result=0;
 	int forward_set=0;
 	int destination_set=0;
         int c, cmd, parse;
-#ifdef HAVE_POPT
         int read_stdin=0;
         int write_stdout=0;
 	dynamic_array_t *a;
@@ -206,59 +205,59 @@ int process_options(int argc, char **argv, int reading_stdin)
 	char *optarg=NULL;
 	
         struct poptOption add_service_option =
-                {"add-service", 'A', POPT_ARG_NONE, NULL, 'A'};
+        {"add-service", 'A', POPT_ARG_NONE, NULL, 'A'};
         struct poptOption edit_service_option =
-                {"edit-service", 'E', POPT_ARG_NONE, NULL, 'E'};
+        {"edit-service", 'E', POPT_ARG_NONE, NULL, 'E'};
         struct poptOption delete_service_option =
-                {"delete-service", 'D', POPT_ARG_NONE, NULL, 'D'};
+        {"delete-service", 'D', POPT_ARG_NONE, NULL, 'D'};
         struct poptOption clear_option =
-                {"clear", 'C', POPT_ARG_NONE, NULL, 'C'};
+        {"clear", 'C', POPT_ARG_NONE, NULL, 'C'};
         struct poptOption list_option =
-                {"list", 'L', POPT_ARG_NONE, NULL, 'L'};
+        {"list", 'L', POPT_ARG_NONE, NULL, 'L'};
         struct poptOption list2_option =
-                {"list", 'l', POPT_ARG_NONE, NULL, 'l'};
+        {"list", 'l', POPT_ARG_NONE, NULL, 'l'};
         struct poptOption add_server_option =
-                {"add-server", 'a', POPT_ARG_NONE, NULL, 'a'};
+        {"add-server", 'a', POPT_ARG_NONE, NULL, 'a'};
         struct poptOption edit_server_option =
-                {"edit-server", 'e', POPT_ARG_NONE, NULL, 'e'};
+        {"edit-server", 'e', POPT_ARG_NONE, NULL, 'e'};
         struct poptOption delete_server_option =
-                {"delete-server", 'd', POPT_ARG_NONE, NULL, 'd'};
+        {"delete-server", 'd', POPT_ARG_NONE, NULL, 'd'};
         struct poptOption help_option =
-                {"help", 'h', POPT_ARG_NONE, NULL, 'h'};
+        {"help", 'h', POPT_ARG_NONE, NULL, 'h'};
         struct poptOption read_stdin_option =
-                {"restore", 'R', POPT_ARG_NONE, NULL, 'R'};
+        {"restore", 'R', POPT_ARG_NONE, NULL, 'R'};
         struct poptOption write_stdout_option =
-                {"save", 'S', POPT_ARG_NONE, NULL, 'S'};
+        {"save", 'S', POPT_ARG_NONE, NULL, 'S'};
         struct poptOption tcp_service_option =
-                {"tcp-service", 't', POPT_ARG_STRING, &optarg, 't'};
+        {"tcp-service", 't', POPT_ARG_STRING, &optarg, 't'};
         struct poptOption udp_service_option =
-                {"udp-service", 'u', POPT_ARG_STRING, &optarg, 'u'};
+        {"udp-service", 'u', POPT_ARG_STRING, &optarg, 'u'};
         struct poptOption fwmark_service_option =
-                {"fwmark-service", 'f', POPT_ARG_STRING, &optarg, 'f'};
+        {"fwmark-service", 'f', POPT_ARG_STRING, &optarg, 'f'};
         struct poptOption scheduler_option =
-                {"scheduler", 's', POPT_ARG_STRING, &optarg, 's'};
+        {"scheduler", 's', POPT_ARG_STRING, &optarg, 's'};
         struct poptOption persistent_option =
-                {"persistent", 'p', POPT_ARG_NONE, NULL, 'p'};
+        {"persistent", 'p', POPT_ARG_NONE, NULL, 'p'};
         struct poptOption netmask_option =
-                {"netmask", 'M', POPT_ARG_STRING, &optarg, 'M'};
+        {"netmask", 'M', POPT_ARG_STRING, &optarg, 'M'};
         struct poptOption real_server_option =
-                {"real-server", 'r', POPT_ARG_STRING, &optarg, 'r'};
+        {"real-server", 'r', POPT_ARG_STRING, &optarg, 'r'};
         struct poptOption real_server2_option =
-                {"real-server", 'R', POPT_ARG_STRING, &optarg, 'R'};
+        {"real-server", 'R', POPT_ARG_STRING, &optarg, 'R'};
         struct poptOption masquerading_option =
-                {"masquerading", 'm', POPT_ARG_NONE, NULL, 'm'};
+        {"masquerading", 'm', POPT_ARG_NONE, NULL, 'm'};
         struct poptOption ipip_option =
-                {"ipip", 'i', POPT_ARG_NONE, NULL, 'i'};
+        {"ipip", 'i', POPT_ARG_NONE, NULL, 'i'};
         struct poptOption gatewaying_option =
-                {"gatewaying", 'g', POPT_ARG_NONE, NULL, 'g'};
+        {"gatewaying", 'g', POPT_ARG_NONE, NULL, 'g'};
         struct poptOption weight_option =
-                {"weight", 'w', POPT_ARG_STRING, &optarg, 'w'};
+        {"weight", 'w', POPT_ARG_STRING, &optarg, 'w'};
         struct poptOption numeric_option =
-                {"numeric", 'n', POPT_ARG_NONE, NULL, 'n'};
+        {"numeric", 'n', POPT_ARG_NONE, NULL, 'n'};
         struct poptOption connection_option =
-                {"connection", 'c', POPT_ARG_NONE, NULL, 'c'};
+        {"connection", 'c', POPT_ARG_NONE, NULL, 'c'};
         struct poptOption NULL_option =
-                {NULL, 0, 0, NULL, 0};
+        {NULL, 0, 0, NULL, 0};
 
 	struct poptOption *options_sub=NULL;
 
@@ -340,7 +339,240 @@ int process_options(int argc, char **argv, int reading_stdin)
 
         if ((cmd = poptGetNextOpt(context)) < 0)
                 usage_exit(argv[0], -1);
-#else
+
+        switch (cmd) {
+        case 'A':	
+                *op = IP_VS_SO_SET_ADD;
+		options_sub = options_service;
+                break;
+        case 'E':	
+                *op = IP_VS_SO_SET_EDIT;
+		options_sub = options_service;
+                break;
+        case 'D':
+                *op = IP_VS_SO_SET_DEL;
+		options_sub = options_delete_service;
+                break;
+        case 'a':
+                *op = IP_VS_SO_SET_ADDDEST;
+		options_sub = options_server;
+                break;
+        case 'e':
+                *op = IP_VS_SO_SET_EDITDEST;
+		options_sub = options_server;
+                break;
+        case 'd':
+                *op = IP_VS_SO_SET_DELDEST;
+		options_sub=options_delete_server;
+                break;
+        case 'C':
+                *op = IP_VS_SO_SET_FLUSH;
+		options_sub=options_NULL;
+                break;
+        case 'L':
+        case 'l':
+                *op = IP_VS_SO_SET_LIST;
+		options_sub=options_list;
+                break;
+	case 'R':
+		read_stdin=1;
+		options_sub=options_NULL;
+		break;
+	case 'S':
+		write_stdout=1;
+		options_sub=options_list;
+		break;
+	case 'h':
+                usage_exit(argv[0], 0);
+		break;
+        default:
+                usage_exit(argv[0], -1);
+        }
+
+	poptFreeContext(context);
+	context = poptGetContext("ipvsadm", argc, (const char **)argv,
+                                 options_sub, 0);
+	/* 
+	 * Mangle the first argument
+	 * The first option from this argument has been read,
+	 * but there may be others
+	 */
+	c = strlen(argv[1]);
+	if (c > 2) {
+                if(argv[1][1] != '-'){
+                        /* Suffle first option out of argument */
+                        memmove(argv[1]+1, argv[1]+2, c-2);   
+                        argv[1][c-1]='\0';
+                }
+        } else {
+                /* Skip argument */
+                poptGetNextOpt(context);
+        }
+
+        while ((c=poptGetNextOpt(context)) >= 0){
+                switch (c) {
+                case 't':
+                case 'u':
+                        if (ur->vfwmark != 0)
+                                fail(2, "fwmark already specified");
+                        if (ur->protocol != 0)
+                                fail(2, "protocol already specified");
+                        ur->protocol =
+                                (c=='t' ? IPPROTO_TCP : IPPROTO_UDP);
+                        parse = parse_service(optarg,
+				              ur->protocol,
+				              &ur->vaddr, 
+				              &ur->vport);
+                        if (!(parse & SERVICE_ADDR )) 
+				fail(2, "illegal virtual server "
+                                     "address[:port] specified");
+                        break;
+                case 'f':
+                        if (ur->vfwmark != 0)
+                                fail(2, "fwmark already specified");
+                        if (ur->protocol != 0)
+                                fail(2, "protocol already specified");
+                        /* 
+                         * Set prtocol to a sane values, even
+                         * though it is not used 
+                         */
+                        ur->protocol = IPPROTO_TCP;
+                        /* 
+                         * Get the fwmark 
+                         */
+                        parse = parse_fwmark(optarg, &ur->vfwmark);
+                        if (parse == 0 || ur->vfwmark == 0 ) 
+				fail(2, "illegal virtual server "
+                                     "fwmark specified");
+                        break;
+                case 's':
+                        if (strlen(ur->sched_name) != 0)
+                                fail(2, "multiple scheduling modules specified");
+                        strncpy(ur->sched_name, optarg, IP_VS_SCHEDNAME_MAXLEN);
+                        break;
+                case 'p':
+                        ur->vs_flags = IP_VS_SVC_F_PERSISTENT;
+                        break;
+                case 'M':
+                        parse = parse_netmask(optarg,
+                                              &ur->netmask);
+                        if (parse != 1)
+                                fail(2, "illegal virtual server "
+                                     "persistent mask specified");
+                        break;
+                case 'r':
+                case 'R':
+                        if (destination_set)
+                                fail(2, "Destination already set");
+                        destination_set=1;
+                        parse = parse_service(optarg,
+                                              ur->protocol,
+                                              &ur->daddr, 
+                                              &ur->dport);
+                        if (!(parse&SERVICE_ADDR)) 
+				fail(2, "illegal real server "
+                                     "address[:port] specified");
+                        /* copy vport to dport if none specified */
+                        if (parse == 1)
+                                ur->dport = ur->vport;
+                        break;
+                case 'i':
+			if(forward_set)
+                        	fail(2, "multiple forward mechanims set");
+			forward_set=1;
+                        ur->conn_flags = IP_VS_CONN_F_TUNNEL;
+                        break;
+                case 'g':
+			if(forward_set)
+                        	fail(2, "multiple forward mechanims set");
+			forward_set=1;
+                        ur->conn_flags = IP_VS_CONN_F_DROUTE;
+                        break;
+                case 'm':
+			if(forward_set)
+                        	fail(2, "multiple forward mechanims set");
+			forward_set=1;
+                        ur->conn_flags = 0;
+                        break;
+                case 'w':
+                        if (ur->weight != -1)
+                                fail(2, "multiple server weights specified");
+                        if ((ur->weight=
+                             string_to_number(optarg,0,65535)) == -1)
+                                fail(2, "illegal weight specified");
+                        break;
+                case 'c':
+                        *options |= OPT_CONNECTION;
+                        break;
+                case 'n':
+                        *options |= OPT_NUMERIC;
+                        *format |= FMT_NUMERIC;
+                        break;
+                default:
+                        fail(2, "invalid option");
+                }
+        }
+
+	if (c < -1) {
+		/* an error occurred during option processing */
+		fprintf(stderr, "%s: %s\n",
+			poptBadOption(context, POPT_BADOPTION_NOALIAS),
+			poptStrerror(c));
+		poptFreeContext(context);
+		return -1;
+	}
+
+	if (read_stdin) {
+                /* avoid infinite loop */
+		if (reading_stdin != 0)
+                	usage_exit(argv[0], -1);
+
+		while ((a = config_stream_read(stdin, argv[0])) != NULL) {
+			int i;
+			if ((i = (int)dynamic_array_get_count(a)) > 1)
+				result = process_options
+                                        (i,
+                                         (char **)dynamic_array_get_vector(a),
+                                         1);
+			dynamic_array_destroy(a, DESTROY_STR);
+		}
+		poptFreeContext(context);
+		return(result);
+	}
+
+        if (write_stdout) {
+                *format |= FMT_RULE;
+                list_vs(*format);
+		poptFreeContext(context);
+                return 0;
+        }
+        
+	/* 
+         * If popt is used then optional arguments (persistent timeout)
+	 * has to be handled last. This has the interesting
+	 * side effect that the first non-option argument will
+	 * be used as the timeout, regardless of its position
+	 * in the argument list
+	 */
+	if (ur->vs_flags == IP_VS_SVC_F_PERSISTENT){
+		optarg = (char*)poptGetArg(context);
+       		parse = parse_timeout(optarg, &ur->timeout);
+       		if (parse == 0)
+                        fail(2, "illegal timeout for persistent service");
+	}
+        poptFreeContext(context);
+        return 0;
+}
+        
+#else /* HAVE_POPT */
+
+int parse_options(int argc, char **argv, int reading_stdin,
+                  struct ip_vs_rule_user *ur, int *op,
+                  unsigned int *options, unsigned int *format)
+{
+	int forward_set=0;
+	int destination_set=0;
+        int c, cmd, parse;
         char *optstr = "";
 
 	struct option long_options[] =
@@ -380,88 +612,43 @@ int process_options(int argc, char **argv, int reading_stdin)
 	optind = 1;
 
 	if ((cmd = getopt_long(argc, argv, "AEDCaedlLh",
-		long_options, NULL)) == EOF)
+                               long_options, NULL)) == EOF)
 		usage_exit(argv[0], -1);
-#endif
-
-        memset(&rule, 0, sizeof(struct ip_vs_rule_user));
 
         switch (cmd) {
         case 'A':	
-                op = IP_VS_SO_SET_ADD;
-#ifdef HAVE_POPT
-		options_sub = options_service;
-#else
+                *op = IP_VS_SO_SET_ADD;
                 optstr = "t:u:f:s:M:p::";
-#endif
                 break;
         case 'E':	
-                op = IP_VS_SO_SET_EDIT;
-#ifdef HAVE_POPT
-		options_sub = options_service;
-#else
+                *op = IP_VS_SO_SET_EDIT;
                 optstr = "t:u:f:s:M:p::";
-#endif
                 break;
         case 'D':
-                op = IP_VS_SO_SET_DEL;
-#ifdef HAVE_POPT
-		options_sub = options_delete_service;
-#else
+                *op = IP_VS_SO_SET_DEL;
                 optstr = "t:u:f:";
-#endif
                 break;
         case 'a':
-                op = IP_VS_SO_SET_ADDDEST;
-#ifdef HAVE_POPT
-		options_sub = options_server;
-#else
+                *op = IP_VS_SO_SET_ADDDEST;
                 optstr = "t:u:f:w:r:R:gmi";
-#endif
                 break;
         case 'e':
-                op = IP_VS_SO_SET_EDITDEST;
-#ifdef HAVE_POPT
-		options_sub = options_server;
-#else
+                *op = IP_VS_SO_SET_EDITDEST;
                 optstr = "t:u:f:w:r:R:gmi";
-#endif
                 break;
         case 'd':
-                op = IP_VS_SO_SET_DELDEST;
-#ifdef HAVE_POPT
-		options_sub=options_delete_server;
-#else
+                *op = IP_VS_SO_SET_DELDEST;
                 optstr = "t:u:f:w:r:R:";
-#endif
                 break;
         case 'C':
-                op = IP_VS_SO_SET_FLUSH;
-#ifdef HAVE_POPT
-		options_sub=options_NULL;
-#else
+                *op = IP_VS_SO_SET_FLUSH;
                 optstr = "";
-#endif
                 break;
         case 'L':
         case 'l':
-                op = IP_VS_SO_SET_LIST;
-#ifdef HAVE_POPT
-		options_sub=options_list;
-#else
+                *op = IP_VS_SO_SET_LIST;
                 optstr = "cn";
-#endif
                 break;
-#ifdef HAVE_POPT
-	case 'R':
-		read_stdin=1;
-		options_sub=options_NULL;
-		break;
-	case 'S':
-		write_stdout=1;
-		options_sub=options_list;
-		break;
-#endif
 	case 'h':
                 usage_exit(argv[0], 0);
 		break;
@@ -469,111 +656,65 @@ int process_options(int argc, char **argv, int reading_stdin)
                 usage_exit(argv[0], -1);
         }
 
-        /*
-         * weight=0 is allowed, which means that server is quiesced.
-         */
-        rule.weight = -1;
-        
-        /*
-         * Set direct routing as default forwarding method
-         */
-        rule.conn_flags = IP_VS_CONN_F_DROUTE;
-
-        /*
-         * Set the default persistent granularity to /32 masking
-         */
-        rule.netmask	= ((u_int32_t) 0xffffffff);
-
-	/* 
-	 * Set the fwmark to 0 (unused)
-	 */
-        rule.vfwmark = 0;
-
-#ifdef HAVE_POPT
-	poptFreeContext(context);
-	context = poptGetContext("ipvsadm", argc, (const char **)argv,
-                                 options_sub, 0);
-	/* 
-	 * Mangle the first argument
-	 * The first option from this argument has been read,
-	 * but there may be others
-	 */
-	c = strlen(argv[1]);
-	if (c > 2) {
-                if(argv[1][1] != '-'){
-                        /* Suffle first option out of argument */
-                        memmove(argv[1]+1, argv[1]+2, c-2);   
-                        argv[1][c-1]='\0';
-                }
-        } else {
-                /* Skip argument */
-                poptGetNextOpt(context);
-        }
-
-        while ((c=poptGetNextOpt(context)) >= 0){
-#else
 	while ((c=getopt_long(argc, argv, optstr,
                               long_options, NULL)) != EOF) {
-#endif
                 switch (c) {
                 case 't':
                 case 'u':
-                        if (rule.vfwmark != 0)
+                        if (ur->vfwmark != 0)
                                 fail(2, "fwmark already specified");
-                        if (rule.protocol != 0)
+                        if (ur->protocol != 0)
                                 fail(2, "protocol already specified");
-                        rule.protocol =
+                        ur->protocol =
                                 (c=='t' ? IPPROTO_TCP : IPPROTO_UDP);
                         parse = parse_service(optarg,
-				              rule.protocol,
-				              &rule.vaddr, 
-				              &rule.vport);
+				              ur->protocol,
+				              &ur->vaddr, 
+				              &ur->vport);
                         if (!(parse & SERVICE_ADDR )) 
 				fail(2, "illegal virtual server "
-                                        "address[:port] specified");
+                                     "address[:port] specified");
                         break;
                 case 'f':
-                        if (rule.vfwmark != 0)
+                        if (ur->vfwmark != 0)
                                 fail(2, "fwmark already specified");
-                        if (rule.protocol != 0)
+                        if (ur->protocol != 0)
                                 fail(2, "protocol already specified");
                         /* 
                          * Set prtocol to a sane values, even
                          * though it is not used 
                          */
-                        rule.protocol = IPPROTO_TCP;
+                        ur->protocol = IPPROTO_TCP;
                         /* 
                          * Get the fwmark 
                          */
-                        parse = parse_fwmark(optarg, &rule.vfwmark);
-                        if (parse == 0 || rule.vfwmark == 0 ) 
+                        parse = parse_fwmark(optarg, &ur->vfwmark);
+                        if (parse == 0 || ur->vfwmark == 0 ) 
 				fail(2, "illegal virtual server "
-                                        "fwmark specified");
+                                     "fwmark specified");
                         break;
                 case 's':
-                        if (strlen(rule.sched_name) != 0)
+                        if (strlen(ur->sched_name) != 0)
                                 fail(2, "multiple scheduling modules specified");
-                        strncpy(rule.sched_name, optarg, IP_VS_SCHEDNAME_MAXLEN);
+                        strncpy(ur->sched_name, optarg, IP_VS_SCHEDNAME_MAXLEN);
                         break;
                 case 'p':
-                        rule.vs_flags = IP_VS_SVC_F_PERSISTENT;
-#ifndef HAVE_POPT
+                        ur->vs_flags = IP_VS_SVC_F_PERSISTENT;
                         if (!optarg && optind < argc && argv[optind][0] != '-'
                             && argv[optind][0] != '!')
                                 optarg = argv[optind++];
                         parse = parse_timeout(optarg,
-                                              &rule.timeout);
+                                              &ur->timeout);
                         if (parse == 0)
                                 fail(2, "illegal timeout "
-                                        "for persistent service");
-#endif
+                                     "for persistent service");
                         break;
                 case 'M':
                         parse = parse_netmask(optarg,
-                                              &rule.netmask);
+                                              &ur->netmask);
                         if (parse != 1)
                                 fail(2, "illegal virtual server "
-                                        "persistent mask specified");
+                                     "persistent mask specified");
                         break;
                 case 'r':
                 case 'R':
@@ -581,113 +722,86 @@ int process_options(int argc, char **argv, int reading_stdin)
                                 fail(2, "Destination already set");
                         destination_set=1;
                         parse = parse_service(optarg,
-                                              rule.protocol,
-                                              &rule.daddr, 
-                                              &rule.dport);
+                                              ur->protocol,
+                                              &ur->daddr, 
+                                              &ur->dport);
                         if (!(parse&SERVICE_ADDR)) 
 				fail(2, "illegal real server "
-                                        "address[:port] specified");
+                                     "address[:port] specified");
                         /* copy vport to dport if none specified */
                         if (parse == 1)
-                                rule.dport = rule.vport;
+                                ur->dport = ur->vport;
                         break;
                 case 'i':
 			if(forward_set)
                         	fail(2, "multiple forward mechanims set");
 			forward_set=1;
-                        rule.conn_flags = IP_VS_CONN_F_TUNNEL;
+                        ur->conn_flags = IP_VS_CONN_F_TUNNEL;
                         break;
                 case 'g':
 			if(forward_set)
                         	fail(2, "multiple forward mechanims set");
 			forward_set=1;
-                        rule.conn_flags = IP_VS_CONN_F_DROUTE;
+                        ur->conn_flags = IP_VS_CONN_F_DROUTE;
                         break;
                 case 'm':
 			if(forward_set)
                         	fail(2, "multiple forward mechanims set");
 			forward_set=1;
-                        rule.conn_flags = 0;
+                        ur->conn_flags = 0;
                         break;
                 case 'w':
-                        if (rule.weight != -1)
+                        if (ur->weight != -1)
                                 fail(2, "multiple server weights specified");
-                        if ((rule.weight=
+                        if ((ur->weight =
                              string_to_number(optarg,0,65535)) == -1)
                                 fail(2, "illegal weight specified");
                         break;
                 case 'c':
-                        options |= OPT_CONNECTION;
+                        *options |= OPT_CONNECTION;
                         break;
                 case 'n':
-                        options |= OPT_NUMERIC;
-                        format |= FMT_NUMERIC;
+                        *options |= OPT_NUMERIC;
+                        *format |= FMT_NUMERIC;
                         break;
                 default:
                         fail(2, "invalid option");
                 }
         }
 
-#ifdef HAVE_POPT
-	if (c < -1) {
-		/* an error occurred during option processing */
-		fprintf(stderr, "%s: %s\n",
-			poptBadOption(context, POPT_BADOPTION_NOALIAS),
-			poptStrerror(c));
-		poptFreeContext(context);
-		return -1;
-	}
-
-	if (read_stdin) {
-                /* avoid infinite loop */
-		if (reading_stdin != 0)
-                	usage_exit(argv[0], -1);
-
-		while ((a = config_stream_read(stdin, argv[0])) != NULL) {
-			int i;
-			if ((i = (int)dynamic_array_get_count(a)) > 1)
-				result = process_options(i,
-                                         (char **)dynamic_array_get_vector(a),
-                                         1);
-			dynamic_array_destroy(a, DESTROY_STR);
-		}
-		poptFreeContext(context);
-		return(result);
-	}
-
-        if (write_stdout) {
-                format |= FMT_RULE;
-                list_vs(format);
-		poptFreeContext(context);
-                return 0;
-        }
-        
-	/* 
-         * If popt is used then optional arguments (persistent timeout)
-	 * has to be handled last. This has the interesting
-	 * side effect that the first non-option argument will
-	 * be used as the timeout, regardless of its position
-	 * in the argument list
-	 */
-	if (rule.vs_flags == IP_VS_SVC_F_PERSISTENT){
-		optarg = (char*)poptGetArg(context);
-       		parse = parse_timeout(optarg, &rule.timeout);
-       		if (parse == 0)
-                        fail(2, "illegal timeout for persistent service");
-	}
-#else
         if (optind < argc)
                 fail(2, "unknown arguments found in command line");
-#endif
+        return 0;
+}
 
+#endif /* HAVE_POPT */
+
+
+int process_options(int argc, char **argv, int reading_stdin)
+{
+        struct ip_vs_rule_user urule;
+        int op=IP_VS_SO_SET_NONE;
+        unsigned int options=OPT_NONE;
+        unsigned int format=FMT_NONE;
+        int sockfd, result;
+
+        memset(&urule, 0, sizeof(struct ip_vs_rule_user));
+        /* weight=0 is allowed, which means that server is quiesced */
+        urule.weight = -1;
+        /* Set direct routing as default forwarding method */
+        urule.conn_flags = IP_VS_CONN_F_DROUTE;
+        /* Set the default persistent granularity to /32 masking */
+        urule.netmask	= ((u_int32_t) 0xffffffff);
+
+        if (parse_options(argc, argv, reading_stdin,
+                          &urule, &op, &options, &format))
+                return -1;
+        
         if (op == IP_VS_SO_SET_LIST) {
                 if (options & OPT_CONNECTION)
                         list_conn(format);
                 else
                         list_vs(format);
-#ifdef HAVE_POPT
-		poptFreeContext(context);
-#endif
 		return 0;
 	}
 
@@ -695,16 +809,16 @@ int process_options(int argc, char **argv, int reading_stdin)
                 /*
                  * Make sure that port zero service is persistent
                  */
-                if (!rule.vfwmark && !rule.vport &&
-                    (rule.vs_flags != IP_VS_SVC_F_PERSISTENT))
+                if (!urule.vfwmark && !urule.vport &&
+                    (urule.vs_flags != IP_VS_SVC_F_PERSISTENT))
                         fail(2, "Zero port specified "
                              "for non-persistent service");
 
                 /*
                  * Set the default scheduling algorithm if not specified
                  */
-                if (strlen(rule.sched_name) == 0)
-                        strcpy(rule.sched_name,DEF_SCHED);
+                if (strlen(urule.sched_name) == 0)
+                        strcpy(urule.sched_name,DEF_SCHED);
         }
 
         /* 
@@ -714,7 +828,7 @@ int process_options(int argc, char **argv, int reading_stdin)
         if ((op == IP_VS_SO_SET_ADDDEST
              || op == IP_VS_SO_SET_EDITDEST
              || op == IP_VS_SO_SET_DELDEST)
-            && !destination_set) {
+            && !urule.daddr) {
                 fail(2, "No destination specified");
         }
 
@@ -722,18 +836,18 @@ int process_options(int argc, char **argv, int reading_stdin)
                 /*
                  * Set the default weight 1 if not specified
                  */
-                if (rule.weight == -1)
-                        rule.weight = 1;
+                if (urule.weight == -1)
+                        urule.weight = 1;
 
                 /*
                  * The destination port must be equal to the service port
                  * if the IP_VS_CONN_F_TUNNEL or IP_VS_CONN_F_DROUTE is set.
                  * Don't worry about this if fwmark is used.
                  */
-                if (!rule.vfwmark &&
-                    (rule.conn_flags == IP_VS_CONN_F_TUNNEL
-                     || rule.conn_flags == IP_VS_CONN_F_DROUTE))
-                        rule.dport = rule.vport;
+                if (!urule.vfwmark &&
+                    (urule.conn_flags == IP_VS_CONN_F_TUNNEL
+                     || urule.conn_flags == IP_VS_CONN_F_DROUTE))
+                        urule.dport = urule.vport;
         }
 
         sockfd = socket(AF_INET, SOCK_RAW, IPPROTO_RAW);
@@ -743,45 +857,40 @@ int process_options(int argc, char **argv, int reading_stdin)
         }
 
         result = setsockopt(sockfd, IPPROTO_IP, op,
-                            (char *)&rule, sizeof(rule));
+                            (char *)&urule, sizeof(urule));
         if (result) {
                 perror("setsockopt failed");
 
                 /*
                  *    Print most common error messages
                  */
-                switch (cmd) {
-                case 'A':	
+                switch (op) {
+                case IP_VS_SO_SET_ADD:	
                         if (errno == EEXIST)
                                 printf("Service already exists\n");
                         else if (errno == ENOENT)
                                 printf("Scheduler not found: ip_vs_%s.o\n",
-                                       rule.sched_name);
+                                       urule.sched_name);
                         break;
-                case 'E':
+                case IP_VS_SO_SET_EDIT:
                         if (errno==ESRCH)
                                 printf("No such service\n");
                         else if (errno == ENOENT)
                                 printf("Scheduler not found: ip_vs_%s.o\n",
-                                       rule.sched_name);
+                                       urule.sched_name);
                         break;
-                case 'D':
+                case IP_VS_SO_SET_DEL:
                         if (errno==ESRCH)
                                 printf("No such service\n");
                         break;
-                case 'a':
+                case IP_VS_SO_SET_ADDDEST:
                         if (errno == ESRCH)
                                 printf("Service not defined\n");
                         else if (errno == EEXIST)
                                 printf("Destination already exists\n");
                         break;
-                case 'e':
-                        if (errno==ESRCH)
-                                printf("Service not defined\n");
-                        else if (errno == ENOENT)
-                                printf("No such destination\n");
-                        break;
-                case 'd':
+                case IP_VS_SO_SET_EDITDEST:
+                case IP_VS_SO_SET_DELDEST:
                         if (errno==ESRCH)
                                 printf("Service not defined\n");
                         else if (errno == ENOENT)
@@ -791,10 +900,6 @@ int process_options(int argc, char **argv, int reading_stdin)
         }
 
         close(sockfd);
-
-#ifdef HAVE_POPT
-	poptFreeContext(context);
-#endif
         return result;	
 }
 
@@ -1075,9 +1180,9 @@ void print_conn(char *buf, unsigned int format)
         vaddr.s_addr = (__u32) htonl(temp2);
         daddr.s_addr = (__u32) htonl(temp3);
 
-        if (!(cname=addrport_to_anyname(&caddr, dport, proto, format)))
+        if (!(cname=addrport_to_anyname(&caddr, cport, proto, format)))
                 exit(1);
-        if (!(vname=addrport_to_anyname(&vaddr, dport, proto, format)))
+        if (!(vname=addrport_to_anyname(&vaddr, vport, proto, format)))
                 exit(1);
         if (!(dname=addrport_to_anyname(&daddr, dport, proto, format)))
                 exit(1);
@@ -1086,7 +1191,7 @@ void print_conn(char *buf, unsigned int format)
 	seconds = (expires / HZ) % 60;
 	minutes = expires / (60 * HZ);
 
-        printf("%s %02ld:%02ld.%02ld %-11s %-17s %-17s %-17s\n",
+        printf("%-3s %02ld:%02ld.%02ld %-11s %-17s %-17s %-17s\n",
                protocol, minutes, seconds, sec100s, state,
                cname, vname, dname);
 
